import re

from capybara.helpers import declension, desc, failure_message


class Result(object):
    """
    A :class:`Result` represents a collection of :class:`Element` objects on the page. It is
    possible to interact with this collection similar to a List because it implements
    ``__getitem__`` and offers the following container methods through delegation:

    * ``__len__``
    * ``__nonzero__`` (Python 2)

    Args:
        elements (List[Element]): The initial list of elements found by the query.
        query (SelectorQuery): The query used to find elements.
    """

    def __init__(self, elements, query):

        # Further reduce the result set using the query's filters.
        self.result = [node for node in elements
                       if query.matches_filters(node)]
        self.rest = list(set(elements) - set(self.result))
        self.query = query

    def __getitem__(self, key):
        return self.result[key]

    def __len__(self):
        return len(self.result)

    def __nonzero__(self):
        return len(self.result) > 0

    @property
    def failure_message(self):
        """ str: A message describing the query failure. """

        message = failure_message(self.query.description, self.query.options)

        if len(self) > 0:
            message += ", found {count} {matches}: {results}".format(
                count=len(self),
                matches=declension("match", "matches", len(self)),
                results=", ".join([desc(node.text) for node in self]))
        else:
            message += " but there were no matches"

        if self.rest:
            elements = ", ".join([desc(element.text) for element in self.rest])
            message += (". Also found {}, which matched the selector"
                        " but not all filters.".format(elements))

        return message

    @property
    def negative_failure_message(self):
        return re.sub(r"(to find)", r"not \1", self.failure_message)



from copy import copy
import pytest

import capybara
from capybara.server import Server
from capybara.tests.app import app


class TestCurrentURL:
    @pytest.fixture(scope="module")
    def servers(self):
        return [Server(copy(app)).boot() for _ in range(2)]

    @pytest.fixture
    def bases(self, servers):
        return ["http://{0}:{1}".format(s.host, s.port) for s in servers]

    @staticmethod
    def visit_host_links(session, bases):
        url = "{0}/host_links?absolute_host={1}".format(bases[0], bases[1])
        session.visit(url)

    @staticmethod
    def assert_on_server(session, server, path="/host", scheme="http"):
        url = "{0}://{1}:{2}{3}".format(scheme, server.host, server.port, path)
        host = "{0}://{1}".format(scheme, server.host)

        assert session.current_url.rstrip("?") == url
        assert session.current_host == host
        assert session.current_path == path

        if path == "/host":
            text = "Current host is {0}://{1}:{2}".format(scheme, server.host, server.port)
            assert session.has_text(text)

    def test_affected_by_visiting_a_page_directly(self, session, servers, bases):
        session.visit("{0}/host".format(bases[0]))
        self.assert_on_server(session, servers[0])

    def test_returns_to_the_app_host_when_visiting_a_relative_url(self, session, servers, bases):
        capybara.app_host = bases[1]

        session.visit("{0}/host".format(bases[0]))
        self.assert_on_server(session, servers[0])

        session.visit("/host")
        self.assert_on_server(session, servers[1])

        capybara.app_host = None

    def test_affected_by_setting_capybara_app_host(self, session, servers, bases):
        capybara.app_host = bases[0]
        session.visit("/host")
        self.assert_on_server(session, servers[0])

        capybara.app_host = bases[1]
        session.visit("/host")
        self.assert_on_server(session, servers[1])

        capybara.app_host = None

    def test_not_affected_by_following_a_relative_link(self, session, servers, bases):
        self.visit_host_links(session, bases)
        session.click_link("Relative Host")
        self.assert_on_server(session, servers[0])

    def test_affected_by_following_an_absolute_link(self, session, servers, bases):
        self.visit_host_links(session, bases)
        session.click_link("Absolute Host")
        self.assert_on_server(session, servers[1])

    def test_not_affected_by_posting_through_a_relative_form(self, session, servers, bases):
        self.visit_host_links(session, bases)
        session.click_button("Relative Host")
        self.assert_on_server(session, servers[0])

    def test_affected_by_posting_through_an_absolute_form(self, session, servers, bases):
        self.visit_host_links(session, bases)
        session.click_link("Absolute Host")
        self.assert_on_server(session, servers[1])

    def test_affected_by_following_a_redirect(self, session, servers, bases):
        session.visit("{0}/redirect".format(bases[0]))
        self.assert_on_server(session, servers[0], path="/landed")

    @pytest.mark.requires("js")
    def test_affected_by_push_state(self, session):
        session.visit("/with_js")
        session.execute_script("window.history.pushState({}, '', '/pushed')")
        assert session.current_path == "/pushed"

    @pytest.mark.requires("js")
    def test_affected_by_replace_state(self, session):
        session.visit("/with_js")
        session.execute_script("window.history.replaceState({}, '', '/replaced')")
        assert session.current_path == "/replaced"



from contextlib import contextmanager
import signal
from socket import socket
from time import time

from capybara.compat import (
    ParseResult,
    bytes_,
    bytes_decode,
    parse_qsl,
    quote,
    str_encode,
    unquote,
    urlencode,
    urlparse)


_missing = object()


class cached_property(property):
    """ Decorates an instance method, turning it into a property whose value is cached. """

    def __init__(self, func):
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def decode_bytes(value):
    """ str: Decodes the given byte sequence. """
    return bytes_decode(value, "utf-8") if isinstance(value, bytes_) else value


def encode_string(value):
    """ bytes: Encodes the given string. """
    return str_encode(value, "utf-8") if isinstance(value, str) else value


def find_available_port():
    """ int: A random available port. """
    s = socket()
    s.bind(("", 0))
    return s.getsockname()[1]


def inner_content(node):
    """
    Returns the inner content of a given XML node, including tags.

    Args:
        node (lxml.etree.Element): The node whose inner content is desired.

    Returns:
        str: The inner content of the node.
    """

    from lxml import etree

    # Include text content at the start of the node.
    parts = [node.text]

    for child in node.getchildren():
        # Include the child serialized to raw XML.
        parts.append(etree.tostring(child, encoding="utf-8"))

        # Include any text following the child.
        parts.append(child.tail)

    # Discard any non-existent text parts and return.
    return "".join(filter(None, parts))


def inner_text(node):
    """
    Returns the inner text of a given XML node, excluding tags.

    Args:
        node: (lxml.etree.Element): The node whose inner text is desired.

    Returns:
        str: The inner text of the node.
    """

    from lxml import etree

    # Include text content at the start of the node.
    parts = [node.text]

    for child in node.getchildren():
        # Include the raw text content of the child.
        parts.append(etree.tostring(child, encoding="utf-8", method="text"))

        # Include any text following the child.
        parts.append(child.tail)

    # Discard any non-existent text parts and return.
    return "".join(map(decode_bytes, filter(None, parts)))


def isbytes(value):
    """ bool: Whether the given value is a sequence of bytes. """
    return isinstance(value, bytes_)


def isregex(possible_regex):
    """
    Returns whether the given object is (probably) a regular expression object.

    Args:
        possible_regex (object): An object which may or may not be a regular expression.

    Returns:
        bool: Whether the given object is (probably) a regular expression object.
    """

    return hasattr(possible_regex, "search") and callable(possible_regex.search)


def normalize_url(url):
    """
    Returns the given URL with all query keys properly escaped.

    Args:
        url (str): The URL to normalize.

    Returns:
        str: The normalized URL.
    """

    uri = urlparse(url)
    query = uri.query or ""

    pairs = parse_qsl(query)
    decoded_pairs = [(unquote(key), value) for key, value in pairs]
    encoded_pairs = [(quote(key), value) for key, value in decoded_pairs]
    normalized_query = urlencode(encoded_pairs)

    return ParseResult(
        scheme=uri.scheme,
        netloc=uri.netloc,
        path=uri.path,
        params=uri.params,
        query=normalized_query,
        fragment=uri.fragment).geturl()


def setter_decorator(fset):
    """
    Define a write-only property that, in addition to the given setter function, also
    provides a setter decorator defined as the property's getter function.

    This allows one to set the property either through traditional assignment, as a
    method argument, or through decoration::

        class Widget(object):
            @setter_decorator
            def handler(self, value):
                self._handler = value

        widget = Widget()

        # Method 1: Traditional assignment
        widget.handler = lambda input: process(input)

        # Method 2: Assignment via method argument
        widget.handler(lambda input: process(input))

        # Method 3: Assignment via decoration
        @widget.handler
        def handler(input):
            return process(input)

        # Method 3b: Assignment via decoration with extraneous parens
        @widget.handler()
        def handler(input):
            return process(input)
    """

    def fget(self):
        def inner(value):
            fset(self, value)

        def outer(value=None):
            if value:
                # We are being called with the desired value, either directly or
                # as a decorator.
                inner(value)
            else:
                # Assume we are being called as a decorator with extraneous parens,
                # so return the setter as the actual decorator.
                return inner

        return outer

    fdoc = fset.__doc__

    return property(fget, fset, None, fdoc)


class TimeoutError(Exception):
    pass


@contextmanager
def timeout(seconds):
    """
    Raises an exception if the wrapped code fails to return within the given number of seconds.

    Args:
        seconds (int | float): The number of seconds to wait before timing out.

    Raises:
        TimeoutError: The wrapped code failed to return within the given timeout.
    """

    def handler(signum, frame):
        raise TimeoutError()

    old_seconds, start_time = 0, 0
    old_handler = signal.signal(signal.SIGALRM, handler)
    try:
        start_time = time()
        old_seconds = signal.alarm(seconds)

        if 0 < old_seconds < seconds:
            # Someone else cares about a shorter timeout, so restore it.
            handler = None

            # Give ourselves a moment to restore before re-activating the alarm.
            signal.alarm(0)

            # Restore the original handler.
            signal.signal(signal.SIGALRM, old_handler)
            old_handler = None

            # Restore the original alarm.
            signal.alarm(old_seconds)
            old_seconds = 0

        try:
            yield
        finally:
            if handler:
                signal.alarm(0)
    finally:
        # If another handler was registered, restore it.
        if old_handler:
            signal.signal(signal.SIGALRM, old_handler)

        # If another alarm was already scheduled, restore it.
        if old_seconds:
            elapsed = time() - start_time
            signal.alarm(old_seconds - elapsed)



import pytest

from capybara.exceptions import ElementNotFound


class SelectorTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestLabelSelector(SelectorTestCase):
    def test_finds_a_label_by_text(self, session):
        assert session.find("label", "Customer Name").text == "Customer Name"

    def test_finds_a_label_by_for_attribute_string(self, session):
        assert session.find("label", field="form_other_title")["for"] == "form_other_title"

    def test_finds_a_label_from_nested_input_using_field_filter(self, session):
        field = session.find("id", "nested_label")
        assert session.find("label", field=field).text == "Nested Label"

    def test_finds_the_label_for_a_non_nested_element_using_field_filter(self, session):
        field = session.find("id", "form_other_title")
        assert session.find("label", field=field)["for"] == "form_other_title"

    def test_matches_substrings_when_exact_is_false(self, session):
        assert session.find("label", "Customer Na", exact=False).text == "Customer Name"

    def test_does_not_match_substrings_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find("label", "Customer Na", exact=True)



from capybara.compat import urlparse
from capybara.helpers import desc
from capybara.utils import isregex, normalize_url


class CurrentPathQuery(object):
    """
    Queries the current path.

    Args:
        expected_path (str | RegexObject): The expected path.
        url (bool, optional): Whether the complete URL should match. Defaults to False.
        only_path (bool, optional): Whether only the path (excluding the query string) should
            match. Defaults to False.
    """

    def __init__(self, expected_path, url=False, only_path=False):
        if url and only_path:
            raise RuntimeError("the url and only_path arguments cannot both be true")

        self.expected_path = expected_path
        self.actual_path = None
        self.url = url
        self.only_path = only_path

    def resolves_for(self, session):
        """
        Returns whether this query resolves for the given session.

        Args:
            session (Session): The session for which this query should be executed.

        Returns:
            bool: Whether this query resolves.
        """

        if self.url:
            self.actual_path = session.current_url
        else:
            result = urlparse(session.current_url)

            if self.only_path:
                self.actual_path = result.path
            else:
                request_uri = result.path
                if result.query:
                    request_uri += "?{0}".format(result.query)

                self.actual_path = request_uri

        if isregex(self.expected_path):
            return self.expected_path.search(self.actual_path)
        else:
            return normalize_url(self.actual_path) == normalize_url(self.expected_path)

    @property
    def failure_message(self):
        """ str: A description of this query's failure. """
        return self._build_message()

    @property
    def negative_failure_message(self):
        """ str: A description of this query's negative failure. """
        return self._build_message(" not")

    def _build_message(self, negated=""):
        verb = "match" if isregex(self.expected_path) else "equal"

        return "expected {actual}{negated} to {verb} {expected}".format(
            actual=desc(self.actual_path),
            negated=negated,
            verb=verb,
            expected=desc(self.expected_path))



import pytest
from xpath import html

import capybara
from capybara.exceptions import ExpectationNotMet


class FindAllTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")


class TestFindAll(FindAllTestCase):
    def test_finds_all_elements_using_the_given_locator(self, session):
        assert len(session.find_all("//p")) == 3
        assert session.find_all("//h1")[0].text == "This is a test"
        assert session.find_all("//input[@id='test_field']")[0].value == "monkey"

    def test_returns_an_empty_result_when_nothing_was_found(self, session):
        assert len(session.find_all("//div[@id='nosuchthing']")) == 0

    def test_accepts_an_xpath_expression_instance(self, session):
        session.visit("/form")
        xpath = html.fillable_field("Name")
        result = [r.value for r in session.find_all(xpath)]
        assert "Smith" in result
        assert "John" in result
        assert "John Smith" in result

    def test_finds_all_elements_using_the_given_css_selector(self, session):
        assert session.find_all("css", "h1")[0].text == "This is a test"
        assert session.find_all("css", "input[id='test_field']")[0].value == "monkey"

    def test_finds_all_elements_when_given_a_list_of_css_selectors(self, session):
        assert len(session.find_all("css", "h1, p")) == 4

    def test_finds_all_elements_using_the_given_xpath_query(self, session):
        assert session.find_all("xpath", "//h1")[0].text == "This is a test"
        assert session.find_all("xpath", "//input[@id='test_field']")[0].value == "monkey"

    def test_finds_the_first_element_using_the_given_locator_when_css_is_the_default_selector(self, session):
        capybara.default_selector = "css"
        assert session.find_all("h1")[0].text == "This is a test"
        assert session.find_all("input[id='test_field']")[0].value == "monkey"

    def test_finds_any_element_using_the_given_locator_in_a_scope(self, session):
        session.visit("/with_scope")
        with session.scope("xpath", "//div[@id='for_bar']"):
            assert len(session.find_all(".//li")) == 2


class TestFindAllVisible(FindAllTestCase):
    def test_only_finds_visible_nodes_when_true(self, session):
        assert len(session.find_all("css", "a.simple", visible=True)) == 1

    def test_finds_nodes_regardless_of_whether_they_are_invisible_when_false(self, session):
        assert len(session.find_all("css", "a.simple", visible=False)) == 2

    def test_defaults_to_capybara_ignore_hidden_elements(self, session):
        capybara.ignore_hidden_elements = True
        assert len(session.find_all("css", "a.simple")) == 1
        capybara.ignore_hidden_elements = False
        assert len(session.find_all("css", "a.simple")) == 2


class TestFindAllCount(FindAllTestCase):
    def test_succeeds_when_the_number_of_elements_found_matches_the_expectation(self, session):
        session.find_all("css", "h1, p", count=4)

    def test_raises_when_the_number_of_elements_found_does_not_match_the_expectation(self, session):
        with pytest.raises(ExpectationNotMet):
            session.find_all("css", "h1, p", count=5)


class TestFindAllMinimum(FindAllTestCase):
    def test_succeeds_when_the_number_of_elements_found_matches_the_expectation(self, session):
        session.find_all("css", "h1, p", minimum=0)

    def test_raises_when_the_number_of_elements_found_does_not_match_the_expectation(self, session):
        with pytest.raises(ExpectationNotMet):
            session.find_all("css", "h1, p", minimum=5)


class TestFindAllMaximum(FindAllTestCase):
    def test_succeeds_when_the_number_of_elements_found_matches_the_expectation(self, session):
        session.find_all("css", "h1, p", maximum=4)

    def test_raises_when_the_number_of_elements_found_does_not_match_the_expectation(self, session):
        with pytest.raises(ExpectationNotMet):
            session.find_all("css", "h1, p", maximum=0)


class TestFindAllBetween(FindAllTestCase):
    def test_succeeds_when_the_number_of_elements_found_matches_the_expectation(self, session):
        session.find_all("css", "h1, p", between=range(2, 7))

    def test_raises_when_the_number_of_elements_found_does_not_match_the_expectation(self, session):
        with pytest.raises(ExpectationNotMet):
            session.find_all("css", "h1, p", between=range(0, 3))



import re

import capybara
from capybara.helpers import declension, desc, failure_message, normalize_text, toregex
from capybara.utils import isregex


VALID_QUERY_TYPE = ["all", "visible"]


class TextQuery(object):
    """
    Queries for text content in a node.

    If only one argument is provided, it will be assumed to be ``expected_text``.

    Args:
        query_type (str, optional): One of "visible" or "all". Defaults to "visible".
        expected_text (str | RegexObject): The desired text.
        between (Iterable[int], optional): A range of acceptable counts.
        count (int, optional): The number of times the text should match. Defaults to any number of
            times greater than zero.
        maximum (int, optional): The maximum number of times the selector should match. Defaults to
            infinite.
        minimum (int, optional): The minimum number of times the selector should match. Defaults to
            1.
        wait (bool | int | float, optional): Whether and how long to wait for synchronization.
            Defaults to :data:`capybara.default_max_wait_time`.
    """

    def __init__(self, query_type, expected_text=None, between=None, count=None, maximum=None,
                 minimum=None, wait=None):
        if expected_text is None:
            expected_text = query_type
            query_type = None

        if query_type is None:
            query_type = ("visible" if capybara.ignore_hidden_elements or capybara.visible_text_only
                          else "all")

        assert query_type in VALID_QUERY_TYPE, \
            "invalid option {query_type} for query_type, should be one of {valid_values}".format(
                query_type=desc(query_type),
                valid_values=", ".join(desc(VALID_QUERY_TYPE)))

        self.expected_text = (expected_text if isregex(expected_text)
                              else normalize_text(expected_text))
        self.query_type = query_type
        self.search_regexp = toregex(expected_text)
        self.options = {
            "between": between,
            "count": count,
            "maximum": maximum,
            "minimum": minimum,
            "wait": wait}
        self.node = None
        self.actual_text = None
        self.count = None

    @property
    def wait(self):
        """ int | float: How long to wait for synchronization. """
        if self.options["wait"] is not None:
            return self.options["wait"] or 0
        else:
            return capybara.default_max_wait_time

    def resolve_for(self, node):
        """
        Resolves this query relative to the given node.

        Args:
            node (node.Base): The node to be evaluated.

        Returns:
            int: The number of matches found.
        """

        self.node = node
        self.actual_text = normalize_text(
            node.visible_text if self.query_type == "visible" else node.all_text)
        self.count = len(re.findall(self.search_regexp, self.actual_text))

        return self.count

    @property
    def description(self):
        if isregex(self.expected_text):
            return "text matching {}".format(desc(self.expected_text))
        else:
            return "text {}".format(desc(self.expected_text))

    @property
    def failure_message(self):
        """ str: A message describing the query failure. """
        return self._build_message(True)

    @property
    def negative_failure_message(self):
        """ str: A message describing the negative query failure. """
        return re.sub(r"(to find)", r"not \1", self._build_message(False))

    def _build_message(self, report_on_invisible):
        message = failure_message(self.description, self.options)
        if any(self.options.values()):
            message += " but found {count} {times}".format(
                count=self.count,
                times=declension("time", "times", self.count))
        message += " in {actual}".format(actual=desc(self.actual_text))

        details_message = []

        if self.node and not self.search_regexp.flags & re.IGNORECASE:
            insensitive_regex = re.compile(
                self.search_regexp.pattern,
                flags=(self.search_regexp.flags | re.IGNORECASE))
            insensitive_count = len(re.findall(insensitive_regex, self.actual_text))
            if insensitive_count != self.count:
                details_message.append(
                    "it was found {count} {times} using a case insensitive search".format(
                        count=insensitive_count,
                        times=declension("time", "times", insensitive_count)))

        if self.node and self.query_type == "visible" and report_on_invisible:
            invisible_text = normalize_text(self.node.all_text)
            invisible_count = len(re.findall(self.search_regexp, invisible_text))
            if invisible_count != self.count:
                details_message.append(
                    "it was found {count} {times} including non-visible text".format(
                        count=invisible_count,
                        times=declension("time", "times", invisible_count)))

        if details_message:
            message += ". (However, {details}.)".format(details=" and ".join(details_message))

        return message



import pytest

import capybara
from capybara.exceptions import Ambiguous, ElementNotFound


class FindFieldTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestFindField(FindFieldTestCase):
    def test_finds_any_field(self, session):
        assert session.find_field("Dog").value == "dog"
        assert session.find_field("form_description").text == "Descriptive text goes here"
        assert session.find_field("Region")["name"] == "form[region]"

    def test_raises_an_error_if_the_field_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.find_field("Does not exist")

    def test_warns_if_filter_option_is_invalid(self, session):
        with pytest.warns(None) as record:
            session.find_field("Dog", disabled=None)
        assert len(record) == 1
        assert "Invalid value None passed to filter disabled - defaulting to False" in record[0].message.args[0]

    def test_raises_an_error_if_the_field_is_disabled(self, session):
        with pytest.raises(ElementNotFound):
            session.find_field("Disabled Checkbox")

    def test_finds_an_approximately_matching_field(self, session):
        assert session.find_field("Explanation")["name"] == "form[name_explanation]"

    def test_does_not_find_an_approximately_matching_field_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find_field("Explanation", exact=True)

    def test_finds_by_aria_label_attribute_when_enable_aria_label_is_true(self, session):
        capybara.enable_aria_label = True
        assert session.find_field("Unlabelled Input")["name"] == "form[which_form]"

    def test_does_not_find_by_aria_label_attribute_when_enable_aria_label_is_false(self, session):
        capybara.enable_aria_label = False
        with pytest.raises(ElementNotFound):
            session.find_field("Unlabelled Input")


class TestFindFieldDisabled(FindFieldTestCase):
    def test_does_not_find_disabled_fields_when_false(self, session):
        with pytest.raises(ElementNotFound):
            session.find_field("Disabled Checkbox", disabled=False)

    def test_finds_enabled_fields_when_false(self, session):
        field = session.find_field("Dog", disabled=False)
        assert field.value == "dog"

    def test_finds_disabled_fields_when_true(self, session):
        field = session.find_field("Disabled Checkbox", disabled=True)
        assert field["name"] == "form[disabled_checkbox]"

    def test_does_not_find_enabled_fields_when_true(self, session):
        with pytest.raises(ElementNotFound):
            session.find_field("Dog", disabled=True)

    def test_finds_disabled_fields_when_all(self, session):
        field = session.find_field("Disabled Checkbox", disabled="all")
        assert field["name"] == "form[disabled_checkbox]"

    def test_finds_enabled_fields_when_all(self, session):
        field = session.find_field("Dog", disabled="all")
        assert field.value == "dog"


class TestFindFieldReadonly(FindFieldTestCase):
    def test_finds_readonly_fields_when_true(self, session):
        assert session.find_field("form[readonly_test]", readonly=True)["id"] == "readonly"

    def test_does_not_find_readonly_fields_when_false(self, session):
        assert session.find_field("form[readonly_test]", readonly=False)["id"] == "not_readonly"

    def test_ignores_readonly_by_default(self, session):
        with pytest.raises(Ambiguous) as excinfo:
            session.find_field("form[readonly_test]")
        assert "found 2 elements" in str(excinfo.value)



import pytest


class HasButtonTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestHasButton(HasButtonTestCase):
    def test_is_true_if_the_given_button_is_on_the_page(self, session):
        assert session.has_button("med")
        assert session.has_button("crap321")

    def test_is_false_if_the_given_button_is_not_on_the_page(self, session):
        assert not session.has_button("monkey")

    def test_is_false_if_the_given_button_is_disabled(self, session):
        assert not session.has_button("Disabled button")


class TestHasButtonDisabled(HasButtonTestCase):
    def test_is_false_for_disabled_buttons_when_false(self, session):
        assert not session.has_button("Disabled button", disabled=False)

    def test_is_true_for_enabled_buttons_when_false(self, session):
        assert session.has_button("med", disabled=False)

    def test_is_true_for_disabled_buttons_when_true(self, session):
        assert session.has_button("Disabled button", disabled=True)

    def test_is_false_for_enabled_buttons_when_true(self, session):
        assert not session.has_button("med", disabled=True)

    def test_is_true_for_disabled_buttons_when_all(self, session):
        assert session.has_button("Disabled button", disabled="all")

    def test_is_true_for_enabled_buttons_when_all(self, session):
        assert session.has_button("med", disabled="all")


class TestHasNoButton:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")

    def test_is_false_if_the_given_button_is_on_the_page(self, session):
        assert not session.has_no_button("med")
        assert not session.has_no_button("crap321")

    def test_is_false_for_disabled_buttons_if_disabled_is_true(self, session):
        assert not session.has_no_button("Disabled button", disabled=True)

    def test_is_true_if_the_given_button_is_not_on_the_page(self, session):
        assert session.has_no_button("monkey")

    def test_is_true_for_disabled_buttons_by_default(self, session):
        assert session.has_no_button("Disabled button")

    def test_is_true_for_disabled_buttons_if_disabled_is_false(self, session):
        assert session.has_no_button("Disabled button", disabled=False)



from lxml import etree
import re

from capybara.utils import inner_content


class HTML(object):
    def __init__(self, source):
        if not source:
            source = "<html/>"

        parser = etree.HTMLParser(encoding="utf-8")
        tree = etree.HTML(source, parser=parser)

        for element in tree.xpath("//textarea"):
            content = inner_content(element)
            content = re.sub("\A\n", "", content)
            for child in element.getchildren():
                element.remove(child)
            element.text = content

        self.tree = tree

    def xpath(self, xpath):
        return self.tree.xpath(xpath)



class CapybaraError(Exception):
    pass


class FrozenInTime(CapybaraError):
    pass


class ElementNotFound(CapybaraError):
    pass


class ModalNotFound(CapybaraError):
    pass


class Ambiguous(ElementNotFound):
    pass


class ExpectationNotMet(ElementNotFound):
    pass


class FileNotFound(CapybaraError):
    pass


class UnselectNotAllowed(CapybaraError):
    pass


class ScopeError(CapybaraError):
    pass


class WindowError(CapybaraError):
    pass


class ReadOnlyElementError(CapybaraError):
    pass



import pytest


class TestHasTable:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/tables")

    def test_is_true_if_the_table_is_on_the_page(self, session):
        assert session.has_table("Villain")
        assert session.has_table("villain_table")

    def test_is_false_if_the_table_is_not_on_the_page(self, session):
        assert not session.has_table("Monkey")


class TestHasNoTable:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/tables")

    def test_is_false_if_the_table_is_on_the_page(self, session):
        assert not session.has_no_table("Villain")
        assert not session.has_no_table("villain_table")

    def test_is_true_if_the_table_is_not_on_the_page(self, session):
        assert session.has_no_table("Monkey")



import capybara
from capybara.exceptions import Ambiguous, ElementNotFound, ExpectationNotMet
from capybara.helpers import matches_count
from capybara.queries.selector_query import SelectorQuery


class FindersMixin(object):
    """
    If the driver is capable of executing JavaScript, finders will wait for a set amount of time and
    continuously retry finding the element until either the element is found or the time expires.
    The length of time :meth:`find` will wait is controlled through
    :data:`capybara.default_max_wait_time` and defaults to 2 seconds.
    """

    def find(self, *args, **kwargs):
        """
        Find an :class:`Element` based on the given arguments. ``find`` will raise an error if the
        element is not found.

        ``find`` takes the same options as :meth:`find_all`. ::

            page.find("#foo").find(".bar")
            page.find("xpath", "//div[contains(., 'bar')]")
            page.find("li", text="Quox").click_link("Delete")

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element.

        Raises:
            Ambiguous: If more than one element matching element is found.
            ElementNotFound: If the element can't be found before time expires.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def find():
            if query.match in ["prefer_exact", "smart"]:
                result = query.resolve_for(self, True)
                if len(result) == 0 and not query.exact:
                    result = query.resolve_for(self, False)
            else:
                result = query.resolve_for(self)

            if query.match in ["one", "smart"] and len(result) > 1:
                raise Ambiguous("Ambiguous match, found {count} elements matching {query}".format(
                    count=len(result), query=query.description))
            if len(result) == 0:
                raise ElementNotFound("Unable to find {0}".format(query.description))

            element = result[0]
            element.allow_reload = True
            return element

        return find()

    def find_button(self, locator, **kwargs):
        """
        Find a button on the page. This can be any ``<input>`` element of type submit, reset, image,
        or button, or it can be a ``<button>`` element. All buttons can be found by their id, value,
        or title. ``<button>`` elements can also be found by their text content, and image
        ``<input>`` elements by their alt attribute.

        Args:
            locator (str): The id, value, title, text content, alt of image.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element.
        """

        return self.find("button", locator, **kwargs)

    def find_by_id(self, id, **kwargs):
        """
        Find an element on the page, given its id.

        Args:
            id (str): The id of the element.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element.
        """

        return self.find("id", id, **kwargs)

    def find_field(self, locator, **kwargs):
        """
        Find a form field on the page. The field can be found by its name, id, or label text.

        Args:
            locator (str): Name, id, placeholder, or text of associated label element.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element.
        """

        return self.find("field", locator, **kwargs)

    def find_link(self, locator, **kwargs):
        """
        Find a link on the page. The link can be found by its id or text.

        Args:
            locator (str): ID, title, text, or alt of enclosed img element.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element.
        """

        return self.find("link", locator, **kwargs)

    def find_all(self, *args, **kwargs):
        """
        Find all elements on the page matching the given selector and options.

        Both XPath and CSS expressions are supported, but Capybara does not try to automatically
        distinguish between them. The following statements are equivalent::

            page.find_all("css", "a#person_123")
            page.find_all("xpath", "//a[@id='person_123']")

        If the type of selector is left out, Capybara uses :data:`capybara.default_selector`. It's
        set to ``"css"`` by default. ::

            page.find_all("a#person_123")

            capybara.default_selector = "xpath"
            page.find_all("//a[@id='person_123']")

        The set of found elements can further be restricted by specifying options. It's possible to
        select elements by their text or visibility::

            page.find_all("a", text="Home")
            page.find_all("#menu li", visible=True)

        By default if no elements are found, an empty list is returned; however, expectations can be
        set on the number of elements to be found which will trigger Capybara's waiting behavior for
        the expectations to match. The expectations can be set using::

            page.assert_selector("p#foo", count=4)
            page.assert_selector("p#foo", maximum=10)
            page.assert_selector("p#foo", minimum=1)
            page.assert_selector("p#foo", between=range(1, 11))

        See :func:`matches_count` for additional information about count matching.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Result: A collection of found elements.

        Raises:
            ExpectationNotMet: The matched results did not meet the expected criteria.
        """

        query = SelectorQuery(*args, **kwargs)

        @self.synchronize(wait=query.wait)
        def find_all():
            result = query.resolve_for(self)

            if not matches_count(len(result), query.options):
                raise ExpectationNotMet(result.failure_message)

            return result

        return find_all()

    def find_first(self, *args, **kwargs):
        """
        Find the first element on the page matching the given selector and options, or None if no
        element matches.

        By default, no waiting behavior occurs. However, if ``capybara.wait_on_first_by_default``
        is set to true, it will trigger Capybara's waiting behavior for a minimum of 1 matching
        element to be found.

        Args:
            *args: Variable length argument list for :class:`SelectorQuery`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Returns:
            Element: The found element or None.
        """

        if capybara.wait_on_first_by_default:
            kwargs.setdefault("minimum", 1)

        try:
            result = self.find_all(*args, **kwargs)
            return result[0] if len(result) > 0 else None
        except ExpectationNotMet:
            return None



import pytest
import re

from capybara.exceptions import ExpectationNotMet


class TestAssertCurrentPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_does_not_raise_if_the_page_has_the_given_current_path(self, session):
        session.assert_current_path("/with_js")

    def test_allows_regexp_matches(self, session):
        session.assert_current_path(re.compile(r"w[a-z]{3}_js"))

    def test_handles_non_escaped_query_options(self, session):
        session.click_link("Non-escaped query options")
        session.assert_current_path("/with_html?options[]=things")

    def test_handles_escaped_query_options(self, session):
        session.click_link("Escaped query options")
        session.assert_current_path("/with_html?options%5B%5D=things")

    @pytest.mark.requires("js")
    def test_waits_for_current_path(self, session):
        session.click_link("Change page")
        session.assert_current_path("/with_html")

    def test_raises_an_error_if_the_page_does_not_have_the_given_current_path(self, session):
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_current_path("/with_html")
        assert "expected '/with_js' to equal '/with_html'" in str(excinfo.value)

    def test_checks_query_options(self, session):
        session.visit("/with_js?test=test")
        session.assert_current_path("/with_js?test=test")

    def test_compares_the_full_url(self, session):
        session.assert_current_path(re.compile(r"\Ahttp://[^/]*/with_js\Z"), url=True)

    def test_ignores_the_query(self, session):
        session.visit("/with_js?test=test")
        session.assert_current_path("/with_js?test=test")
        session.assert_current_path("/with_js", only_path=True)


class TestAssertNoCurrentPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_raises_if_the_page_has_the_given_current_path(self, session):
        with pytest.raises(ExpectationNotMet):
            session.assert_no_current_path("/with_js")

    def test_allows_regex_matches(self, session):
        session.assert_no_current_path(re.compile(r"monkey"))

    @pytest.mark.requires("js")
    def test_waits_for_current_path_to_disappear(self, session):
        session.click_link("Change page")
        session.assert_no_current_path("/with_js")

    def test_does_not_raise_if_the_page_does_not_have_the_given_current_path(self, session):
        session.assert_no_current_path("/with_html")



import pytest

from capybara.exceptions import ModalNotFound


@pytest.mark.requires("modals")
class TestAcceptPrompt:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_accepts_the_prompt_with_no_message(self, session):
        with session.accept_prompt():
            session.click_link("Open prompt")
        assert session.has_xpath("//a[@id='open-prompt' and @response='']")

    def test_accepts_the_prompt_with_a_response(self, session):
        with session.accept_prompt(response="the response"):
            session.click_link("Open prompt")
        assert session.has_xpath("//a[@id='open-prompt' and @response='the response']")

    def test_accepts_the_prompt_if_the_message_matches(self, session):
        with session.accept_prompt("Prompt opened", response="matched"):
            session.click_link("Open prompt")
        assert session.has_xpath("//a[@id='open-prompt' and @response='matched']")

    def test_raises_error_if_the_message_does_not_match(self, session):
        with pytest.raises(ModalNotFound):
            with session.accept_prompt("Incorrect Text", response="not matched"):
                session.click_link("Open prompt")



import pytest

import capybara
from capybara.compat import urlparse
from capybara.session import Session


class TestVisit:
    def test_fetches_a_response_from_the_driver_with_a_relative_url(self, session):
        session.visit("/")
        assert session.has_text("Hello world!")
        session.visit("/foo")
        assert session.has_text("Another World")

    def test_fetches_a_response_from_the_driver_with_an_absolute_url_with_a_port(self, session):
        # Preparation
        session.visit("/")
        root_uri = urlparse(session.current_url)

        session.visit("http://{}/".format(root_uri.netloc))
        assert session.has_text("Hello world!")
        session.visit("http://{}/foo".format(root_uri.netloc))
        assert session.has_text("Another World")

    def test_fetches_a_response_when_absolute_uri_does_not_have_a_trailing_slash(self, session):
        # Preparation
        session.visit("/")
        root_uri = urlparse(session.current_url)

        session.visit("http://{}".format(root_uri.netloc))
        assert session.has_text("Hello world!")

    def test_sets_cookie_if_a_blank_path_is_specified(self, session):
        session.visit("")
        session.visit("/get_cookie")
        assert session.has_text("root cookie")


class TestVisitWithoutApp:
    def test_does_not_instantiate_a_server(self, session):
        serverless_session = Session(session.mode, None)
        assert serverless_session.server is None

    @pytest.mark.requires("server")
    def test_respects_app_host(self, session):
        serverless_session = Session(session.mode, None)
        capybara.app_host = "http://{}:{}".format(session.server.host, session.server.port)
        serverless_session.visit("/foo")
        assert serverless_session.has_text("Another World")

    @pytest.mark.requires("server")
    def test_visits_a_fully_qualified_url(self, session):
        serverless_session = Session(session.mode, None)
        serverless_session.visit(
            "http://{}:{}/foo".format(session.server.host, session.server.port))
        assert serverless_session.has_text("Another World")



import pytest

from capybara.tests.helpers import extract_results


class TestWithFieldset:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/fieldsets")

    def test_restricts_scope_to_a_fieldset_given_by_id(self, session):
        with session.fieldset("villain_fieldset"):
            session.fill_in("Name", value="Goldfinger")
            session.click_button("Create")
        assert extract_results(session)["form[villain_name]"] == "Goldfinger"

    def test_restricts_scope_to_a_fieldset_given_by_legend(self, session):
        with session.fieldset("Villain"):
            session.fill_in("Name", value="Goldfinger")
            session.click_button("Create")
        assert extract_results(session)["form[villain_name]"] == "Goldfinger"



from io import BytesIO
import os
import re
from werkzeug.datastructures import FileStorage, MultiDict
from xpath import dsl as x
from xpath.renderer import to_xpath

from capybara.werkzeug.node import Node


class Form(Node):
    def params(self, button):
        params = MultiDict()

        types = ["input", "select", "textarea"]
        xpath = x.descendant(*types)[~x.attr("form")]
        if self.native.get("id"):
            xpath += x.anywhere(*types)[x.attr("form") == self.native.get("id")]
        xpath = xpath[~x.attr("disabled")]

        for field in self._find_xpath(to_xpath(xpath)):
            if field.tag_name == "input":
                if field["type"] in ["checkbox", "radio"]:
                    if field.checked:
                        params.add(field["name"], field.value)
                elif field["type"] == "file":
                    if self._multipart:
                        if field.value:
                            params.add(
                                field["name"],
                                FileStorage(
                                    stream=open(field.value, "rb"),
                                    filename=os.path.basename(field.value)))
                        else:
                            params.add(
                                field["name"],
                                FileStorage(
                                    stream=BytesIO(),
                                    content_type="application/octet-stream"))
                    else:
                        if field.value:
                            params.add(field["name"], os.path.basename(field.value))
                elif field["type"] in ["submit", "reset", "image"]:
                    pass
                else:
                    params.add(field["name"], field.value)
            elif field.tag_name == "textarea":
                if field.value:
                    params.add(field["name"], re.sub("\n", "\r\n", field.value))
            elif field.tag_name == "select":
                if field["multiple"] == "multiple":
                    options = field.native.xpath(".//option[@selected='selected']")
                    for option in options:
                        params.add(field["name"], option.get("value", option.text))
                else:
                    options = field.native.xpath(".//option[@selected='selected']")
                    if not len(options):
                        options = field.native.xpath(".//option")
                    if len(options):
                        params.add(field["name"], options[0].get("value", options[0].text))

        params.add(button["name"], button["value"] or "")

        return params

    def submit(self, button):
        action = self.native.get("action")
        method = self._request_method
        self.driver.submit(method, action, self.params(button))

    @property
    def _request_method(self):
        if self["method"] and re.compile(r"post", re.IGNORECASE).search(self["method"]):
            return "POST"
        else:
            return "GET"

    @property
    def _multipart(self):
        return self["enctype"] == "multipart/form-data"



import pytest

import capybara
from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class ChooseTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestChoose(ChooseTestCase):
    def test_chooses_a_radio_button_by_id(self, session):
        session.choose("gender_male")
        session.click_button("awesome")
        assert extract_results(session)["form[gender]"] == "male"

    def test_chooses_a_radio_button_by_label(self, session):
        session.choose("Both")
        session.click_button("awesome")
        assert extract_results(session)["form[gender]"] == "both"

    def test_chooses_an_approximately_matching_radio_button(self, session):
        session.choose("Mal")
        session.click_button("awesome")
        assert extract_results(session)["form[gender]"] == "male"

    def test_raises_an_error_with_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.choose("does not exist")

    def test_raises_an_error_for_a_disabled_radio_button(self, session):
        with pytest.raises(ElementNotFound):
            session.choose("Disabled Radio")


class TestExactChoose(ChooseTestCase):
    def test_raises_an_error_for_an_approximately_matching_radio_button(self, session):
        with pytest.raises(ElementNotFound):
            session.choose("Mal", exact=True)


class TestChooseWithAutomaticLabelClick(ChooseTestCase):
    @pytest.fixture(autouse=True)
    def setup_settings(self):
        old_automatic_label_click = capybara.automatic_label_click
        capybara.automatic_label_click = True
        try:
            yield
        finally:
            capybara.automatic_label_click = old_automatic_label_click

    def test_selects_by_clicking_the_link_if_available(self, session):
        session.choose("party_democrat")
        session.click_button("awesome")
        assert extract_results(session)["form[party]"] == "democrat"

    def test_raises_an_error_if_not_allowed_to_click_label(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.choose("party_democrat", allow_label_click=False)
        assert "Unable to find radio button 'party_democrat'" in str(excinfo.value)



import pytest


class TestHasTitle:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_is_true_if_the_page_has_the_given_title(self, session):
        assert session.has_title("with_js")

    @pytest.mark.requires("js")
    def test_waits_for_title(self, session):
        session.click_link("Change title")
        assert session.has_title("changed title")

    def test_is_false_if_the_page_does_not_have_the_given_title(self, session):
        assert not session.has_title("monkey")


class TestHasNoTitle:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_is_false_if_the_page_has_the_given_title(self, session):
        assert session.has_no_title("with_js") is False

    @pytest.mark.requires("js")
    def test_waits_for_title_to_disappear(self, session):
        session.click_link("Change title")
        assert session.has_no_title("with_js") is True

    def test_is_true_if_the_page_does_not_have_the_given_title(self, session):
        assert session.has_no_title("monkey") is True



from contextlib import contextmanager


class Base(object):
    """ The base class for drivers used by sessions. """

    @property
    def needs_server(self):
        """ bool: Whether this driver needs to communicate with a real HTTP server. """
        return False

    @property
    def current_url(self):
        """ str: The current URL. """
        raise NotImplementedError()

    @property
    def title(self):
        """ str: The current page title. """
        raise NotImplementedError()

    @property
    def html(self):
        """ str: A snapshot of the DOM of the current document, as it looks right now. """
        raise NotImplementedError()

    def switch_to_frame(self, frame):
        """
        Switch to the given frame.

        Args:
            frame (Element | str): The element of the desired frame, or "parent".
        """

        raise NotImplementedError()

    @property
    def current_window_handle(self):
        """ object: The current window handle. """
        raise NotImplementedError()

    def window_size(self, handle):
        """ List[int, int]: The size of the given window handle. """
        raise NotImplementedError()

    def resize_window_to(self, handle, width, height):
        """
        Resizes the given window handle to the given dimensions.

        Args:
            handle (object): The driver handle for the window.
            width (int): The desired window width.
            height (int): The desired window height.
        """

        raise NotImplementedError()

    def maximize_window(self, handle):
        """ Maximizes the given window handle. """
        raise NotImplementedError()

    def close_window(self, handle):
        """
        Closes the window with the given handle.

        args:
            handle (object): The handle for the desired window.
        """

        raise NotImplementedError()

    @property
    def window_handles(self):
        """ List[object]: The handles for all opened windows. """
        raise NotImplementedError()

    def open_new_window(self):
        """ Opens a new window. """
        raise NotImplementedError()

    def switch_to_window(self, handle):
        """
        Switches to the given window handle.

        Args:
            handle (object): The handle for the desired window.
        """

        raise NotImplementedError()

    @property
    def no_such_window_error(self):
        """ Exception: The error that is thrown when a window cannot be found. """
        raise NotImplementedError()

    def visit(self, path):
        """
        Visits the given path.

        Args:
            path (str): The path to visit.
        """

        raise NotImplementedError()

    def go_back(self):
        """ Move back a single entry in the browser's history. """
        raise NotImplementedError()

    def go_forward(self):
        """ Move forward a single entry in the browser's history. """
        raise NotImplementedError()

    def execute_script(self, script, *args):
        """
        Executes the given script.

        Args:
            script (str): A JavaScript string to execute.
            *args: Variable length argument list to pass to the executed JavaScript string.
        """

        raise NotImplementedError()

    def evaluate_script(self, script, *args):
        """
        Evaluates and returns the result of the given JavaScript.

        Args:
            script (str): A string of JavaScript to evaluate.
            *args: Variable length argument list to pass to the executed JavaScript string.

        Returns:
            object: The result of the evaluated JavaScript.
        """

        raise NotImplementedError()

    def save_screenshot(self, path, **kwargs):
        """
        Saves a screenshot of the current state of the page.

        Args:
            path (str): The path to where it should be saved.
            **kwargs: Arbitrary keywords arguments for the driver.
        """

        raise NotImplementedError()

    @contextmanager
    def accept_modal(self, modal_type, text=None, response=None, wait=None):
        """
        Accepts the modal that appears matching the given type and, optionally, text.

        Args:
            modal_type (str): The type of modal that should be accepted.
            text (str, optional): Text that is expected to appear in the modal.
            response (str, optional): Text to enter for a response, if applicable.
            wait (int, optional): The number of seconds to wait for the modal to appear.
        """

        raise NotImplementedError()

    @contextmanager
    def dismiss_modal(self, modal_type, text=None, wait=None):
        """
        Dismisses the modal that appears matching the given type and, optionally, text.

        Args:
            modal_type (str): The type of modal that should be dismissed.
            text (str, optional): Text that is expected to appear in the modal.
            wait (int, optional): The number of seconds to wait for the modal to appear.
        """

        raise NotImplementedError()

    @property
    def invalid_element_errors(self):
        """ Tuple[Exception]: A tuple of exceptions that indicate an element is invalid. """
        return ()

    def reset(self):
        """ Resets the driver. """
        pass

    def _find_css(self, query):
        """
        A private method for finding nodes matching a given CSS query.

        Args:
            query (str): The CSS query to match.

        Returns:
            list(driver.Node): A list of matching nodes found by the driver.
        """

        raise NotImplementedError()

    def _find_xpath(self, query):
        """
        A private method for finding nodes matching a given XPath query.

        Args:
            query (str): The XPath query to match.

        Returns:
            List[driver.Node]: A list of matching nodes found by the driver.
        """

        raise NotImplementedError()



# coding=utf-8
from __future__ import unicode_literals

import pytest
import re

import capybara
from capybara.exceptions import ExpectationNotMet


class TestAssertText:
    def test_is_true_if_the_given_text_is_on_the_page(self, session):
        session.visit("/with_html")
        assert session.assert_text("est") is True
        assert session.assert_text("Lorem") is True
        assert session.assert_text("Redirect") is True
        assert session.assert_text("text with whitespace") is True
        assert session.assert_text("text     with \n\n whitespace") is True

    def test_raises_an_error_if_the_given_text_is_not_on_the_page(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet):
            session.assert_text("thisisnotonthepage")

    def test_is_true_if_the_given_unicode_text_is_on_the_page(self, session):
        session.visit("/with_html")
        assert session.assert_text("이름") is True

    def test_raises_an_error_if_the_given_unicode_text_is_not_on_the_page(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet):
            session.assert_text("论坛")

    def test_takes_scopes_into_account(self, session):
        session.visit("/with_html")
        with session.scope("//a[@title='awesome title']"):
            assert session.assert_text("labore") is True

    def test_raises_if_scoped_to_an_element_which_does_not_have_the_text(self, session):
        session.visit("/with_html")
        with session.scope("//a[@title='awesome title']"):
            with pytest.raises(ExpectationNotMet) as excinfo:
                session.assert_text("monkey")
        assert "expected to find text 'monkey' in 'labore'" in str(excinfo.value)

    def test_is_true_if_all_given_and_text_is_invisible(self, session):
        session.visit("/with_html")
        assert session.assert_text("all", "Some of this text is hidden!") is True

    def test_raises_an_error_with_a_helpful_message_if_the_requested_text_is_present_but_invisible(self, session):
        session.visit("/with_html")
        el = session.find("css", "#hidden-text")
        with pytest.raises(ExpectationNotMet) as excinfo:
            el.assert_text("visible", "Some of this text is hidden!")
        assert "it was found 1 time including non-visible text" in str(excinfo.value)

    def test_raises_error_with_a_helpful_message_if_the_requested_text_is_present_but_with_incorrect_case(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_text("Text With Whitespace")
        assert "it was found 1 time using a case insensitive search" in str(excinfo.value)

    def test_is_true_if_the_text_in_the_page_matches_given_regex(self, session):
        session.visit("/with_html")
        assert session.assert_text(re.compile(r"Lorem")) is True

    def test_raises_error_if_the_text_in_the_page_does_not_match_given_regex(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_text(re.compile(r"xxxxyzzz"))
        assert "expected to find text matching 'xxxxyzzz' in 'This is a test Header Class" in str(excinfo.value)

    def test_excapes_any_characters_that_would_have_special_meaning_in_a_regex(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet):
            session.assert_text(".orem")

    @pytest.mark.requires("js")
    def test_waits_for_text_to_appear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.assert_text("Has been clicked") is True

    def test_is_true_if_the_text_occurs_within_the_range_given(self, session):
        session.visit("/with_count")
        assert session.assert_text("count", between=range(1, 4)) is True

    def test_is_false_if_the_text_occurs_more_or_fewer_times_than_range(self, session):
        session.visit("/with_count")
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_text("count", between=range(0, 2))
        assert "expected to find text 'count' between 0 and 1 times but found 2 times" in str(excinfo.value)

    @pytest.mark.requires("js")
    def test_finds_element_if_it_appears_before_given_wait_duration(self, session):
        with capybara.using_wait_time(0):
            session.visit("/with_js")
            session.find("css", "#reload-list").click()
            session.find("css", "#the-list").assert_text("Foo Bar", wait=0.9)

    @pytest.mark.requires("js")
    def test_raises_error_if_it_appears_after_given_wait_duration(self, session):
        with capybara.using_wait_time(0):
            session.visit("/with_js")
            session.find("css", "#reload-list").click()
            el = session.find("css", "#the-list", visible=False)
            with pytest.raises(ExpectationNotMet):
                el.assert_text("all", "Foo Bar", wait=0.3)


class TestAssertTextCount:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_ignores_other_filters_when_count_is_specified(self, session):
        session.assert_text("Header", count=5, minimum=6, maximum=0, between=range(0, 5))

    def test_fails_if_minimum_is_not_met(self, session):
        with pytest.raises(ExpectationNotMet):
            session.assert_text("Header", minimum=0, maximum=0, between=range(2, 8))

    def test_fails_if_maximum_is_not_met(self, session):
        with pytest.raises(ExpectationNotMet):
            session.assert_text("Header", minimum=0, maximum=0, between=range(2, 8))

    def test_fails_if_between_is_not_met(self, session):
        with pytest.raises(ExpectationNotMet):
            session.assert_text("Header", minimum=0, maximum=5, between=range(0, 5))

    def test_succeeds_if_all_combined_expectations_are_met(self, session):
        session.assert_text("Header", minimum=0, maximum=5, between=range(2, 8))


class TestAssertNoText:
    def test_raises_error_if_the_given_text_is_on_the_page_at_least_once(self, session):
        session.visit("/with_html")
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_no_text("Lorem")
        assert "expected not to find text 'Lorem' in 'This is a test Header Class" in str(excinfo.value)

    def test_is_true_if_scoped_to_an_element_which_does_not_have_the_text(self, session):
        session.visit("/with_html")
        with session.scope("//a[@title='awesome title']"):
            assert session.assert_no_text("monkey") is True

    def test_is_true_if_given_text_is_on_the_page_but_not_visible(self, session):
        session.visit("/with_html")
        assert session.assert_no_text("Inside element with hidden selector") is True

    def test_raises_error_if_all_given_and_text_is_invisible(self, session):
        session.visit("/with_html")
        el = session.find("css", "#hidden-text", visible=False)
        with pytest.raises(ExpectationNotMet) as excinfo:
            el.assert_no_text("all", "Some of this text is hidden!")
        assert "expected not to find text 'Some of this text is hidden!' in 'Some of this text is hidden!'" in str(excinfo.value)

    def test_raises_error_if_visible_given_and_text_is_visible(self, session):
        session.visit("/with_html")
        el = session.find("css", "#some-hidden-text", visible=False)
        with pytest.raises(ExpectationNotMet) as excinfo:
            el.assert_no_text("visible", "hidden")
        assert "expected not to find text 'hidden' in 'Some of this text is not hidden'" in str(excinfo.value)

    def test_is_true_if_the_text_in_the_page_does_not_match_given_regex(self, session):
        session.visit("/with_html")
        session.assert_no_text(re.compile(r"xxxxyzzz"))

    def test_is_true_if_the_text_does_not_occur_the_given_number_of_times(self, session):
        session.visit("/with_count")
        assert session.assert_no_text("count", count=3) is True

    def test_raises_if_the_text_occurs_the_given_number_of_times(self, session):
        session.visit("/with_count")
        with pytest.raises(ExpectationNotMet) as excinfo:
            session.assert_no_text("count", count=2)
        assert "expected not to find text 'count' 2 times" in str(excinfo.value)

    @pytest.mark.requires("js")
    def test_does_not_find_text_if_it_appears_after_given_wait_duration(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        session.find("css", "#reload-list").click()
        session.find("css", "#the-list").assert_no_text("Foo Bar", wait=0.3)



import pytest

from capybara.node.element import Element


@pytest.mark.requires("js")
class TestEvaluateScript:
    def test_evaluates_the_given_script_and_returns_whatever_it_produces(self, session):
        session.visit("/with_js")
        assert session.evaluate_script("1+3") == 4

    def test_passes_arguments_to_the_script(self, session):
        session.visit("/with_js")
        session.evaluate_script("document.getElementById('change').textContent = arguments[0]", "Doodle Funk")
        assert session.has_css("#change", text="Doodle Funk")

    def test_supports_passing_elements_as_arguments_to_the_script(self, session):
        session.visit("/with_js")
        el = session.find("css", "#change")
        session.evaluate_script("arguments[1].textContent = arguments[0]", "Doodle Funk", el)
        assert session.has_css("#change", text="Doodle Funk")

    def test_supports_returning_elements(self, session):
        session.visit("/with_js")
        el = session.evaluate_script("document.getElementById('change')")
        assert isinstance(el, Element)
        assert el == session.find("css", "#change")

    def test_supports_returning_arrays_of_elements(self, session):
        session.visit("/form")
        elements = session.evaluate_script("document.querySelectorAll('#form_city option')")
        for element in elements:
            assert isinstance(element, Element)
        assert elements == list(session.find("css", "#form_city").find_all("css", "option"))



import pytest

from capybara.tests.helpers import extract_results


class TestWithTable:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/tables")

    def test_restricts_scope_to_a_table_given_by_id(self, session):
        with session.table("girl_table"):
            session.fill_in("Name", value="Christmas")
            session.click_button("Create")
        assert extract_results(session)["form[girl_name]"] == "Christmas"

    def test_restricts_scope_to_a_table_given_by_caption(self, session):
        with session.table("Villain"):
            session.fill_in("Name", value="Quantum")
            session.click_button("Create")
        assert extract_results(session)["form[villain_name]"] == "Quantum"



from contextlib import contextmanager

from capybara.selector.filter import Filter


filter_sets = {}
# Dict[str, FilterSet]: A dictionary of global filter sets keyed by name.


class FilterSet(object):
    """
    A group of filters for use by :class:`Selector` objects.

    Args:
        name (str): The name of this set of filters.
        descriptions (List[Callable[[Dict[str, Any]], str]]): Functions that generate descriptions
            for options.
        filters (Dict[str, Filter], optional): The filters in this set. Defaults to {}.
    """

    def __init__(self, name, descriptions=None, filters=None):
        self.name = name
        self.descriptions = descriptions or []
        self.filters = filters or {}

    def description(self, options):
        """
        Returns a description of the given filter options relevant to this filter set.

        Args:
            options (Dict[str, Any]): The filter options to describe.

        Returns:
            str: A description of the filter options.
        """

        return "".join([describe(options) for describe in self.descriptions])


class FilterSetFactory(object):
    """ A factory for configuring and building :class:`FilterSet` instances. """

    def __init__(self, name):
        self.name = name
        self.descriptions = []
        self.filters = {}

    def describe(self, func):
        """
        Decorates a function that builds a description of some filter set options.

        Args:
            func (Callable[[Dict[str, Any]], str]): The description builder function.
        """

        self.descriptions.append(func)

    def filter(self, name, **kwargs):
        """
        Returns a decorator function for adding filters to built sets.

        Args:
            name (str): The name of the filter.
            **kwargs: Variable keyword arguments for the filter.

        Returns:
            Callable[[Callable[[Element, Any], bool]]]: A decorator function for adding filters to
                built sets.
        """

        def decorator(func):
            self.filters[name] = Filter(name, func, **kwargs)

        return decorator

    def build_filter_set(self):
        """ FilterSet: Returns a new :class:`FilterSet` with the current factory config. """
        return FilterSet(self.name, self.descriptions, self.filters)


@contextmanager
def add_filter_set(name):
    """
    Builds and registers a global :class:`FilterSet`.

    Args:
        name (str): The name of the set.

    Yields:
        FilterSetFactory: A configurable factory for building a :class:`FilterSet`.
    """

    factory = FilterSetFactory(name)
    yield factory
    filter_sets[name] = factory.build_filter_set()


def remove_filter_set(name):
    """
    Unregisters filter set with the given name.

    Args:
        name (str): The name of the filter set.
    """

    filter_sets.pop(name, None)



from werkzeug.test import Client, create_environ
from werkzeug.wrappers import BaseResponse, Request

import capybara
from capybara.compat import ParseResult, urlparse
from capybara.html import HTML
from capybara.utils import cached_property, decode_bytes


class Browser(object):
    def __init__(self, driver):
        self.driver = driver
        self.last_request = None
        self.last_response = None
        self._current_scheme = None
        self._current_netloc = None
        self._dom = None

    @property
    def app(self):
        return self.driver.app

    @cached_property
    def client(self):
        return Client(self.app, BaseResponse)

    @property
    def current_url(self):
        if self.last_request:
            return self.last_request.url

    def visit(self, path):
        self._reset_host()
        self._process_and_follow_redirects("GET", path)

    def follow(self, method, path, params=None):
        if path.startswith("#") or path.lower().startswith("javascript:"):
            return
        self._process_and_follow_redirects(method, path, params)

    def submit(self, method, path, params):
        self._process_and_follow_redirects(method, path, params)

    @property
    def html(self):
        return decode_bytes(self.last_response.data) if self.last_response else ""

    @property
    def dom(self):
        if self._dom is None:
            self._dom = HTML(self.html)
        return self._dom

    def _process_and_follow_redirects(self, method, path, params=None, headers=None):
        self._process(method, path, params, headers)
        for _ in range(self.driver.redirect_limit):
            if 300 <= self.last_response.status_code < 400:
                headers = headers.copy() if headers else {}
                path = self.last_response.headers["Location"]
                headers["Referer"] = self.last_request.url
                self._process(method, path, params, headers)

    def _process(self, method, path, params=None, headers=None):
        self._dom = None

        requested_uri = urlparse(path)

        base_uri = ParseResult(
            scheme=requested_uri.scheme or self._current_scheme,
            netloc=requested_uri.netloc or self._current_netloc,
            path="",
            params="",
            query="",
            fragment="")

        path_uri = ParseResult(
            scheme="",
            netloc="",
            path=requested_uri.path,
            params=requested_uri.params,
            query=requested_uri.query,
            fragment=requested_uri.fragment)

        self._current_scheme = base_uri.scheme
        self._current_netloc = base_uri.netloc

        env = create_environ(method=method,
                             path=path_uri.geturl(),
                             base_url=base_uri.geturl() or None,
                             data=params,
                             headers=headers)

        self.last_request = Request(env)
        self.last_response = self.client.open(env)

    def _reset_host(self):
        if capybara.app_host:
            uri = urlparse(capybara.app_host)
            self._current_scheme = uri.scheme
            self._current_netloc = uri.netloc



import pytest
import re

import capybara


class TestHasSelector:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_true_if_the_given_selector_is_on_the_page(self, session):
        assert session.has_selector("xpath", "//p")
        assert session.has_selector("css", "p a#foo")
        assert session.has_selector("//p[contains(.,'est')]")

    def test_is_false_if_the_given_selector_is_not_on_the_page(self, session):
        assert not session.has_selector("xpath", "//abbr")
        assert not session.has_selector("css", "p a#doesnotexist")
        assert not session.has_selector("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_default_selector(self, session):
        capybara.default_selector = "css"
        assert not session.has_selector("p a#doesnotexist")
        assert session.has_selector("p a#foo")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert session.has_selector(".//a[@id='foo']")
            assert not session.has_selector(".//a[@id='red']")

    def test_is_true_if_the_content_is_on_the_page_the_given_number_of_times(self, session):
        assert session.has_selector("//p", count=3)
        assert session.has_selector("//p//a[@id='foo']", count=1)
        assert session.has_selector("//p[contains(.,'est')]", count=1)

    def test_is_false_if_the_content_is_not_on_the_page_the_given_number_of_times(self, session):
        assert not session.has_selector("//p", count=6)
        assert not session.has_selector("//p//a[@id='foo']", count=2)
        assert not session.has_selector("//p[contains(.,'est')]", count=5)

    def test_is_false_if_the_content_is_not_on_the_page_at_all(self, session):
        assert not session.has_selector("//abbr", count=2)
        assert not session.has_selector("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_not_contained(self, session):
        assert session.has_selector("//p//a", text="Redirect", count=1)
        assert not session.has_selector("//p", text="Doesnotexist")

    def test_respects_visibility_setting(self, session):
        assert session.has_selector("id", "hidden-text", text="Some of this text is hidden!", visible=False)
        assert not session.has_selector("id", "hidden-text", text="Some of this text is hidden!", visible=True)
        capybara.ignore_hidden_elements = False
        assert session.has_selector("id", "hidden-text", text="Some of this text is hidden!", visible=False)
        capybara.visible_text_only = True
        assert not session.has_selector("id", "hidden-text", text="Some of this text is hidden!", visible=True)

    def test_discards_all_matches_where_the_given_regex_is_not_matched(self, session):
        assert session.has_selector("//p//a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        assert not session.has_selector("//p//a", text=re.compile("Red$"))


class TestHasNoSelector:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_false_if_the_given_selector_is_on_the_page(self, session):
        assert not session.has_no_selector("xpath", "//p")
        assert not session.has_no_selector("css", "p a#foo")
        assert not session.has_no_selector("//p[contains(.,'est')]")

    def test_is_true_if_the_given_selector_is_not_on_the_page(self, session):
        assert session.has_no_selector("xpath", "//abbr")
        assert session.has_no_selector("css", "p a#doesnotexist")
        assert session.has_no_selector("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_default_selector(self, session):
        capybara.default_selector = "css"
        assert session.has_no_selector("p a#doesnotexist")
        assert not session.has_no_selector("p a#foo")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert not session.has_no_selector(".//a[@id='foo']")
            assert session.has_no_selector("../a[@id='red']")

    def test_is_false_if_the_content_is_on_the_page_the_given_number_of_times(self, session):
        assert not session.has_no_selector("//p", count=3)
        assert not session.has_no_selector("//p//a[@id='foo']", count=1)
        assert not session.has_no_selector("//p[contains(.,'est')]", count=1)

    def test_is_true_if_the_content_is_on_the_page_the_wrong_number_of_times(self, session):
        assert session.has_no_selector("//p", count=6)
        assert session.has_no_selector("//p//a[@id='foo']", count=2)
        assert session.has_no_selector("//p[contains(.,'est')]", count=5)

    def test_is_true_if_the_content_is_not_on_the_page_at_all(self, session):
        assert session.has_no_selector("//abbr", count=2)
        assert session.has_no_selector("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_contained(self, session):
        assert not session.has_no_selector("//p//a", text="Redirect", count=1)
        assert session.has_no_selector("//p", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_matched(self, session):
        assert not session.has_no_selector("//p//a", text=re.compile(r"re[dab]i", re.IGNORECASE), count=1)
        assert session.has_no_selector("//p//a", text=re.compile(r"Red$"))



import json
from lxml import etree
from werkzeug.datastructures import ImmutableMultiDict


def extract_results(session):
    session.assert_selector("//pre[@id='results']")

    tree = etree.HTML(session.body)
    element = tree.xpath("//pre[@id='results']")[0]

    def inner_html(elem):
        return "".join(
            [elem.text or ""] +
            [etree.tostring(child).decode("utf-8") for child in elem] +
            [elem.tail or ""])

    return ImmutableMultiDict(json.loads(inner_html(element)))



import pytest
import re

import capybara


class TestHasXPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_true_if_the_given_selector_is_on_the_page(self, session):
        assert session.has_xpath("//p")
        assert session.has_xpath("//p//a[@id='foo']")
        assert session.has_xpath("//p[contains(.,'est')]")

    def test_is_false_if_the_given_selector_is_not_on_the_page(self, session):
        assert not session.has_xpath("//abbr")
        assert not session.has_xpath("//p//a[@id='doesnotexist']")
        assert not session.has_xpath("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_xpath_even_if_default_selector_is_css(self, session):
        capybara.default_selector = "css"
        assert not session.has_xpath("//p//a[@id='doesnotexist']")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert session.has_xpath(".//a[@id='foo']")
            assert not session.has_xpath(".//a[@id='red']")

    @pytest.mark.requires("js")
    def test_waits_for_content_to_appear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_xpath("//input[@type='submit' and @value='New Here']")

    def test_is_true_if_the_content_occurs_the_given_number_of_times(self, session):
        assert session.has_xpath("//p", count=3)
        assert session.has_xpath("//p//a[@id='foo']", count=1)
        assert session.has_xpath("//p[contains(.,'est')]", count=1)
        assert session.has_xpath("//p//a[@id='doesnotexist']", count=0)

    def test_is_false_if_the_content_occurs_a_different_number_of_times_than_the_given(self, session):
        assert not session.has_xpath("//p", count=6)
        assert not session.has_xpath("//p//a[@id='foo']", count=2)
        assert not session.has_xpath("//p[contains(.,'est')]", count=5)
        assert not session.has_xpath("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_not_contained(self, session):
        assert session.has_xpath("//p//a", text="Redirect", count=1)
        assert not session.has_xpath("//p", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_not_matched(self, session):
        assert session.has_xpath("//p//a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        assert not session.has_xpath("//p", text=re.compile("Red$"))


class TestHasNoXPath:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_false_if_the_given_selector_is_on_the_page(self, session):
        assert not session.has_no_xpath("//p")
        assert not session.has_no_xpath("//p//a[@id='foo']")
        assert not session.has_no_xpath("//p[contains(.,'est')]")

    def test_is_true_if_the_given_selector_is_not_on_the_page(self, session):
        assert session.has_no_xpath("//abbr")
        assert session.has_no_xpath("//p//a[@id='doesnotexist']")
        assert session.has_no_xpath("//p[contains(.,'thisstringisnotonpage')]")

    def test_uses_xpath_even_if_default_selector_is_css(self, session):
        capybara.default_selector = "css"
        assert session.has_no_xpath("//p//a[@id='doesnotexist']")

    def test_respects_scopes(self, session):
        with session.scope("//p[@id='first']"):
            assert not session.has_no_xpath(".//a[@id='foo']")
            assert session.has_no_xpath(".//a[@id='red']")

    @pytest.mark.requires("js")
    def test_waits_for_content_to_disappear(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        assert session.has_no_xpath("//p[@id='change']")

    def test_is_false_if_the_content_occurs_the_given_number_of_times(self, session):
        assert not session.has_no_xpath("//p", count=3)
        assert not session.has_no_xpath("//p//a[@id='foo']", count=1)
        assert not session.has_no_xpath("//p[contains(.,'est')]", count=1)
        assert not session.has_no_xpath("//p//a[@id='doesnotexist']", count=0)

    def test_is_true_if_the_content_occurs_a_different_number_of_times_than_the_given(self, session):
        assert session.has_no_xpath("//p", count=6)
        assert session.has_no_xpath("//p//a[@id='foo']", count=2)
        assert session.has_no_xpath("//p[contains(.,'est')]", count=5)
        assert session.has_no_xpath("//p//a[@id='doesnotexist']", count=1)

    def test_discards_all_matches_where_the_given_string_is_contained(self, session):
        assert not session.has_no_xpath("//p//a", text="Redirect", count=1)
        assert session.has_no_xpath("//p", text="Doesnotexist")

    def test_discards_all_matches_where_the_given_regex_is_matched(self, session):
        assert not session.has_no_xpath("//p//a", text=re.compile("re[dab]i", re.IGNORECASE), count=1)
        assert session.has_no_xpath("//p", text=re.compile("Red$"))



from __future__ import absolute_import
from contextlib import contextmanager


app = None
""" object: The WSGI-compliant app to test. """

app_host = None
""" str: The default host to use when giving a relative URL to visit. Must be a valid URL. """

default_driver = "werkzeug"
""" str: The name of the driver default driver to use. """

javascript_driver = "selenium"
""" str: The name of the driver to use when JavaScript is required. """

current_driver = None
""" str: The name of the driver currently in use. """

server_name = "default"
""" str: The name of the server to use to serve the app. """

server_host = "127.0.0.1"
""" str: The IP address bound by the default server. """

server_port = None
""" int, optional: The port bound by the default server. """

automatic_label_click = False
""" bool: Whether checkbox/radio actions will try to click the label of invisible elements. """

automatic_reload = True
""" bool: Whether to automatically reload elements as Capybara is waiting. """

enable_aria_label = False
""" bool: Whether fields, links, and buttons will match against aria-label attributes. """

exact = False
""" bool: Whether to match the exact label name/contents. """

default_max_wait_time = 2
""" int: The maximum number of seconds to wait for asynchronous processes to finish. """

default_selector = "css"
""" str: The name of the default selector used to find elements. """

ignore_hidden_elements = True
""" bool: Whether to ignore hidden elements on the page. """

match = "smart"
""" str: The matching strategy to use. """

save_path = None
""" str, optional: Where to put saved pages and screenshots. """

visible_text_only = False
""" bool: Whether to only consider visible text. """

wait_on_first_by_default = False
""" bool: Whether :meth:`find_first` should wait for at least one element to appear. """

servers = {}
# Dict[str, Callable[[object, str, int], None]]: A dictionary of server initialization functions.

drivers = {}
# Dict[str, Callable[[object], object]]: A dictionary of driver initialization functions.

session_name = "default"
""" str: The current session name. """

_session_pool = {}
# Dict[str, Session]: A pool of `Session` objects, keyed by driver and app.


DSL_METHODS = ["using_session", "using_wait_time"]


def register_server(name):
    """
    Register a server initialization function.

    Args:
        name (str): The name of the server.

    Returns:
        Callable[[Callable[[object, str, int], None]], None]: A decorator that takes a function
            that initializes a server for the given WSGI-compliant app, host, and port.
    """

    def register(init_func):
        servers[name] = init_func

    return register


def register_driver(name):
    """
    Register a driver initialization function.

    Args:
        name (str): The name of the driver.

    Returns:
        Callable[[Callable[[object], object], None]: A decorator that takes a function that
            initializes a driver for the given WSGI-compliant app.
    """

    def register(init_func):
        drivers[name] = init_func

    return register


def run_default_server(app, port):
    servers["werkzeug"](app, port, server_host)


def use_default_driver():
    """ Use the default driver as the current driver. """
    global current_driver
    current_driver = None


@contextmanager
def using_driver(driver):
    """
    Execute the wrapped code using a specific driver.

    Args:
        driver (str): The name of the desired driver.
    """

    global current_driver

    original_current_driver = current_driver
    current_driver = driver
    try:
        yield
    finally:
        current_driver = original_current_driver


@contextmanager
def using_wait_time(seconds):
    """
    Execute a context using a specific wait time.

    Args:
        seconds (int | float): The new wait time.
    """

    global default_max_wait_time

    original_max_wait_time = default_max_wait_time
    default_max_wait_time = seconds
    try:
        yield
    finally:
        default_max_wait_time = original_max_wait_time


def current_session():
    """
    Returns the :class:`Session` for the current driver and app, instantiating one if needed.

    Returns:
        Session: The :class:`Session` for the current driver and app.
    """

    driver = current_driver or default_driver

    session_key = "{driver}:{session}:{app}".format(
        driver=driver, session=session_name, app=str(id(app)))
    session = _session_pool.get(session_key, None)

    if session is None:
        from capybara.session import Session
        session = Session(driver, app)
        _session_pool[session_key] = session

    return session


@contextmanager
def using_session(name):
    """
    Execute the wrapped code using a specific session name.

    Args:
        name (str): The name of the session to use.
    """

    global session_name

    previous_session_name = session_name
    session_name = name
    try:
        yield
    finally:
        session_name = previous_session_name


def reset_sessions():
    """ Resets all sessions. """

    for session in iter(_session_pool.values()):
        session.reset()


reset = reset_sessions
""" Alias for :func:`reset_sessions`. """


def string(html):
    """
    Wraps the given string, which should contain an HTML document or fragment in a :class:`Simple`
    which exposes :class:`MatchersMixin`, :class:`FindersMixin`, and :class:`DocumentMatchersMixin`.
    This allows you to query any string containing HTML in the exact same way you would query the
    current document in a Capybara session.

    Example: A single element ::

        node = Capybara.string(\"""<a href="foo">bar</a>\""")
        anchor = node.find_first("a")
        anchor["href"]  # => "foo"
        anchor.text  # => "bar"

    Example: Multiple elements ::

        node = Capybara.string(\"""
          <ul>
            <li id="home">Home</li>
            <li id="projects">Projects</li>
          </ul>
        \""")

        node.find("#projects").text  # => "Projects"
        node.has_selector("li#home", text="Home")
        node.has_selector("#projects")
        node.find("ul").find("li:first-child").text  # => "Home"

    Args:
        html (str | lxml.etree.Element): An HTML fragment or document.

    Returns:
        Simple: A node which has Capybara's finders and matchers.
    """

    from capybara.node.simple import Simple

    return Simple(html)


@register_server("default")
def init_default_server(app, port, host):
    run_default_server(app, port)


@register_server("werkzeug")
def init_werkzeug_server(app, port, host):
    try:
        import werkzeug
    except ImportError:
        raise ImportError(
            'Capybara\'s werkzeug server is unable to load `werkzeug`, please install the package '
            'and add `werkzeug` to your requirements.txt file.')

    from werkzeug.serving import make_server
    from logging import getLogger

    # Mute the server.
    log = getLogger('werkzeug')
    log.disabled = True

    server = make_server(host, port, app, threaded=True)

    # Inform Python that it shouldn't wait for request threads to terminate before
    # exiting. (They will still be appropriately terminated when the process exits.)
    server.daemon_threads = True

    server.serve_forever()


@register_driver("selenium")
def init_selenium_driver(app):
    try:
        import selenium
    except ImportError:
        raise ImportError(
            'Capybara\'s selenium driver is unable to load `selenium`, please install the package '
            'and add `selenium` to your requirements.txt file.')

    from capybara.selenium.driver import Driver

    return Driver(app)


@register_driver("werkzeug")
def init_werkzeug_driver(app):
    try:
        import werkzeug
    except ImportError:
        raise ImportError(
            'Capybara\'s werkzeug driver is unable to load `werkzeug`, please install the package '
            'and add `werkzeug` to your requirements.txt file.')

    from capybara.werkzeug.driver import Driver

    return Driver(app)



import pytest

from capybara.exceptions import ModalNotFound


@pytest.mark.requires("modals")
class TestAcceptConfirm:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_accepts_the_confirm(self, session):
        with session.accept_confirm():
            session.click_link("Open confirm")
        assert session.has_xpath("//a[@id='open-confirm' and @confirmed='true']")

    def test_accepts_the_confirm_if_the_message_matches(self, session):
        with session.accept_confirm("Confirm opened"):
            session.click_link("Open confirm")
        assert session.has_xpath("//a[@id='open-confirm' and @confirmed='true']")

    def test_raises_an_error_if_the_message_does_not_match(self, session):
        with pytest.raises(ModalNotFound):
            with session.accept_confirm("Incorrect Text"):
                session.click_link("Open confirm")

    def test_works_with_nested_modals(self, session):
        with session.dismiss_confirm("Are you really sure?"):
            with session.accept_confirm("Are you sure?"):
                session.click_link("Open check twice")
        assert session.has_xpath("//a[@id='open-twice' and @confirmed='false']")



from contextlib import contextmanager

from capybara.selector.filter import Filter
from capybara.selector.filter_set import filter_sets
from capybara.utils import setter_decorator


selectors = {}
# Dict[str, Selector]: A dictionary of :class:`Selector` objects keyed by name. """


class Selector(object):
    """
    A callable object used for selecting elements in a document or node.

    Args:
        name (str): The name of the selector.
        label (str, optional): The label to use when describing this selector.
        descriptions (List[Callable[[Dict[str, Any]], str]]): Functions that build a description of
            the given filter options.
        css (Callable[[str], str], optional): A function to generate a CSS selector given a locator
            string.
        xpath (Callable[[str], str], optional): A function to generate an XPath query given a
            locator string.
        custom_filters (Dict[str, Filter]): A dictionary of filters this selector should use
            to identify matching elements. Defaults to {}.
    """

    def __init__(self, name, label=None, descriptions=None, css=None, xpath=None,
                 custom_filters=None):
        self.name = name
        self.label = label
        self.descriptions = descriptions or []
        self.css = css
        self.xpath = xpath
        self.format = "xpath" if xpath else "css"
        self.custom_filters = custom_filters or {}

    def __call__(self, locator):
        assert self.format, "selector has no format"
        return getattr(self, self.format)(locator)

    def description(self, options):
        """
        Returns a description of the given filter options relevant to this selector.

        Args:
            options (Dict[str, Any]): The filter options to describe.

        Returns:
            str: A description of the filter options.
        """

        return "".join([describe(options) for describe in self.descriptions])


class SelectorFactory(object):
    """
    A factory for configuring and building :class:`Selector` instances.

    Args:
        name (str): The name of the selector.
    """

    def __init__(self, name):
        self.name = name
        self.label = None
        self.descriptions = []
        self.func = None
        self.format = None
        self.custom_filters = {}

    def describe(self, func):
        """
        Decorates a function that builds a description of some selector options.

        Args:
            func (Callable[[Dict[str, Any]], str]): The description builder function.
        """

        self.descriptions.append(func)

    @setter_decorator
    def css(self, func):
        """
        Sets the given function as the CSS selector generation function.

        Args:
            func (Callable[[str], str]): The CSS selector generation function.
        """

        self.func = func
        self.format = "css"

    @setter_decorator
    def xpath(self, func):
        """
        Sets the given function as the XPath query generation function.

        Args:
            func (Callable[[str], str]): The XPath query generation function.
        """

        self.func = func
        self.format = "xpath"

    def filter(self, name, **kwargs):
        """
        Returns a decorator function for adding afilter.

        Args:
            name (str): The name of the filter.
            **kwargs: Variable keyword arguments for the filter.

        Returns:
            Callable[[Callable[[Element, Any], bool]]]: A decorator function for adding a filter.
        """

        def decorator(func):
            self.custom_filters[name] = Filter(name, func, **kwargs)

        return decorator

    def filter_set(self, name):
        """
        Adds filters from a particular global :class:`FilterSet`.

        Args:
            name (str): The name of the set whose filters should be added.
        """

        filter_set = filter_sets[name]
        for name, filter_ in iter(filter_set.filters.items()):
            self.custom_filters[name] = filter_
        self.descriptions += filter_set.descriptions

    def build_selector(self):
        """ Selector: Returns a new :class:`Selector` instance with the current configuration. """

        kwargs = {
            'label': self.label,
            'descriptions': self.descriptions,
            'custom_filters': self.custom_filters}
        if self.format == "xpath":
            kwargs['xpath'] = self.func
        if self.format == "css":
            kwargs['css'] = self.func

        return Selector(self.name, **kwargs)


@contextmanager
def add_selector(name):
    """
    Builds and registers a :class:`Selector` object with the given name and configuration.

    Args:
        name (str): The name of the selector.

    Yields:
        SelectorFactory: The factory that will build the :class:`Selector`.
    """

    factory = SelectorFactory(name)
    yield factory
    selectors[name] = factory.build_selector()


def remove_selector(name):
    """
    Unregisters selector with the given name.

    Args:
        name (str): The name of the selector.
    """

    selectors.pop(name, None)



import pytest
from time import sleep

import capybara
from capybara.exceptions import ReadOnlyElementError
from capybara.tests.helpers import extract_results


class NodeTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")


class TestNode(NodeTestCase):
    def test_acts_like_a_session_object(self, session):
        session.visit("/form")
        form = session.find("css", "#get-form")
        assert form.has_field("Middle Name")
        assert form.has_no_field("Languages")
        form.fill_in("Middle Name", value="Monkey")
        form.click_button("med")
        assert extract_results(session)["form[middle_name]"] == "Monkey"

    def test_scopes_css_selectors(self, session):
        assert not session.find("css", "#second").has_css("h1")


class TestNodeQueryScope(NodeTestCase):
    def test_returns_a_reference_to_the_element_the_query_was_evaluated_on(self, session):
        node = session.find("css", "#first")
        assert node.query_scope == node.session.document
        assert node.find("css", "#foo").query_scope == node


class TestNodeText(NodeTestCase):
    def test_extracts_node_text(self, session):
        assert session.find_all("//a")[0].text == "labore"
        assert session.find_all("//a")[1].text == "ullamco"

    def test_returns_document_text_on_html_selector(self, session):
        session.visit("/with_simple_html")
        assert session.find("/html").text == "Bar"


class TestNodeAttribute(NodeTestCase):
    def test_extracts_node_attributes(self, session):
        assert session.find_all("//a")[0]["class"] == "simple"
        assert session.find_all("//a")[1]["id"] == "foo"
        assert session.find_all("//input")[0]["type"] == "text"

    def test_extracts_boolean_node_attributes(self, session):
        assert session.find("//input[@id='checked_field']")["checked"]


class TestNodeValue(NodeTestCase):
    def test_allows_retrieval_of_the_value(self, session):
        assert session.find("//textarea[@id='normal']").value == "banana"

    def test_does_not_swallow_extra_newlines_in_textarea(self, session):
        assert session.find("//textarea[@id='additional_newline']").value == "\nbanana"

    def test_does_not_swallow_newlines_for_set_content_in_textarea(self, session):
        session.find("//textarea[@id='normal']").set("\nbanana")
        assert session.find("//textarea[@id='normal']").value == "\nbanana"

    def test_returns_any_html_content_in_textarea(self, session):
        session.find("//textarea[1]").set("some <em>html</em>here")
        assert session.find("//textarea[1]").value == "some <em>html</em>here"

    def test_defaults_to_on_for_checkboxes(self, session):
        session.visit("/form")
        assert session.find("//input[@id='valueless_checkbox']").value == "on"

    def test_defaults_to_on_for_radio_buttons(self, session):
        session.visit("/form")
        assert session.find("//input[@id='valueless_radio']").value == "on"


class TestNodeSet(NodeTestCase):
    def test_allows_assignment_of_field_value(self, session):
        assert session.find_first("//input").value == "monkey"
        session.find_first("//input").set("gorilla")
        assert session.find_first("//input").value == "gorilla"

    @pytest.mark.requires("js")
    def test_fills_the_field_even_if_the_caret_was_not_at_the_end(self, session):
        session.execute_script(
            "var el = document.getElementById('test_field');"
            "el.focus();"
            "el.setSelectionRange(0, 0);")
        session.find_first("//input").set("")
        assert session.find_first("//input").value == ""

    def test_raises_if_the_text_field_is_readonly(self, session):
        assert session.find_first("//input[@readonly]").value == "should not change"
        with pytest.raises(ReadOnlyElementError):
            session.find_first("//input[@readonly]").set("changed")
        assert session.find_first("//input[@readonly]").value == "should not change"

    def test_raises_if_the_textarea_is_readonly(self, session):
        assert session.find_first("//textarea[@readonly]").value == "textarea should not change"
        with pytest.raises(ReadOnlyElementError):
            session.find_first("//textarea[@readonly]").set("changed")
        assert session.find_first("//textarea[@readonly]").value == "textarea should not change"

    @pytest.mark.requires("js")
    def test_allows_me_to_change_the_contents_of_a_contenteditable_element(self, session):
        session.visit("/with_js")
        session.find("css", "#existing_content_editable").set("WYSIWYG")
        assert session.find("css", "#existing_content_editable").text == "WYSIWYG"

    @pytest.mark.requires("js")
    def test_allows_me_to_set_the_contents_of_a_contenteditable_element(self, session):
        session.visit("/with_js")
        session.find("css", "#blank_content_editable").set("WYSIWYG")
        assert session.find("css", "#blank_content_editable").text == "WYSIWYG"


class TestNodeTagName(NodeTestCase):
    def test_extracts_node_tag_name(self, session):
        assert session.find_all("//a")[0].tag_name == "a"
        assert session.find_all("//a")[1].tag_name == "a"
        assert session.find_all("//p")[0].tag_name == "p"


class TestNodeDisabled(NodeTestCase):
    def test_extracts_disabled_node(self, session):
        session.visit("/form")
        assert session.find("//input[@id='customer_name']").disabled
        assert not session.find("//input[@id='customer_email']").disabled

    def test_sees_disabled_options_as_disabled(self, session):
        session.visit("/form")
        assert not session.find("//select[@id='form_title']/option[1]").disabled
        assert session.find("//select[@id='form_title']/option[@disabled]").disabled

    def test_sees_enabled_options_in_disabled_select_as_disabled(self, session):
        session.visit("/form")
        assert session.find("//select[@id='form_disabled_select']/option").disabled
        assert not session.find("//select[@id='form_title']/option[1]").disabled

    def test_is_boolean(self, session):
        session.visit("/form")
        assert session.find("//select[@id='form_disabled_select']/option").disabled is True
        assert session.find("//select[@id='form_disabled_select2']/option").disabled is True
        assert session.find("//select[@id='form_title']/option[1]").disabled is False


class TestNodeVisible(NodeTestCase):
    def test_extracts_node_visibility(self, session):
        capybara.ignore_hidden_elements = False
        assert session.find_first("//a").visible
        assert not session.find("//div[@id='hidden']").visible
        assert not session.find("//div[@id='hidden_via_ancestor']").visible
        assert not session.find("//div[@id='hidden_attr']").visible
        assert not session.find("//a[@id='hidden_attr_via_ancestor']").visible
        assert not session.find("//input[@id='hidden_input']").visible

    def test_is_boolean(self, session):
        capybara.ignore_hidden_elements = False
        assert session.find_first("//a").visible is True
        assert session.find("//div[@id='hidden']").visible is False


class TestNodeChecked(NodeTestCase):
    def test_extracts_node_checked_state(self, session):
        session.visit("/form")
        assert session.find("//input[@id='gender_female']").checked is True
        assert session.find("//input[@id='gender_male']").checked is False
        assert session.find_first("//h1").checked is False


class TestNodeSelected(NodeTestCase):
    def test_extracts_node_selected_state(self, session):
        session.visit("/form")
        assert session.find("//option[@value='en']").selected is True
        assert session.find("//option[@value='sv']").selected is False
        assert session.find_first("//h1").checked is False


class TestNodeEquals(NodeTestCase):
    def test_is_true_for_the_same_element(self, session):
        assert session.find("//h1") == session.find("//h1")

    def test_is_false_for_different_elements(self, session):
        assert session.find("//h1") != session.find_first("//p")

    def test_is_false_for_unrelated_objects(self, session):
        assert session.find("//h1") != "Not a node"


class TestNodePath(NodeTestCase):
    def test_returns_xpath_which_points_to_itself(self, session):
        session.visit("/path")
        element = session.find("link", "Second Link")
        assert session.find("xpath", element.path) == element


@pytest.mark.requires("js", "drag")
class TestNodeDragTo(NodeTestCase):
    def test_drags_and_drops_an_object(self, session):
        session.visit("/with_js")
        element = session.find("//div[@id='drag']")
        target = session.find("//div[@id='drop']")
        element.drag_to(target)
        assert session.find("//div[contains(., 'Dropped!')]")


@pytest.mark.requires("hover")
class TestNodeHover(NodeTestCase):
    def test_allows_hovering_on_an_element(self, session):
        session.visit("/with_hover")
        assert not session.find("css", ".hidden_until_hover", visible=False).visible
        session.find("css", ".wrapper").hover()
        assert session.find("css", ".hidden_until_hover", visible=False).visible


class TestNodeClick(NodeTestCase):
    def test_does_not_follow_a_link_if_no_href(self, session):
        session.find("css", "#link_placeholder").click()
        assert session.current_url.endswith("/with_html")

    def test_is_able_to_check_a_checkbox(self, session):
        session.visit("/form")
        checkbox = session.find("checkbox", "form_terms_of_use")
        assert not checkbox.checked
        checkbox.click()
        assert checkbox.checked

    def test_is_able_to_uncheck_a_checkbox(self, session):
        session.visit("/form")
        checkbox = session.find("checkbox", "form_pets_dog")
        assert checkbox.checked
        checkbox.click()
        assert not checkbox.checked

    def test_is_able_to_select_a_radio_button(self, session):
        session.visit("/form")
        radio = session.find("radio_button", "gender_male")
        assert not radio.checked
        radio.click()
        assert radio.checked


@pytest.mark.requires("js")
class TestNodeDoubleClick(NodeTestCase):
    def test_double_clicks_an_element(self, session):
        session.visit("/with_js")
        session.find("css", "#click-test").double_click()
        assert session.find("css", "#has-been-double-clicked")


@pytest.mark.requires("js")
class TestNodeRightClick(NodeTestCase):
    def test_right_clicks_an_element(self, session):
        session.visit("/with_js")
        session.find("css", "#click-test").right_click()
        assert session.find("css", "#has-been-right-clicked")


@pytest.mark.requires("send_keys")
class TestNodeSendKeys(NodeTestCase):
    def test_sends_a_string_of_keys_to_an_element(self, session):
        session.visit("/form")
        session.find("css", "#address1_city").send_keys("Oceanside")
        assert session.find("css", "#address1_city").value == "Oceanside"

    def test_sends_special_characters(self, session):
        from selenium.webdriver.common.keys import Keys

        session.visit("/form")
        session.find("css", "#address1_city").send_keys(
            "Ocean", Keys.SPACE, "sie", Keys.LEFT, "d")
        assert session.find("css", "#address1_city").value == "Ocean side"


@pytest.mark.requires("js")
class TestNodeReloadWithoutAutomaticReload(NodeTestCase):
    @pytest.fixture(autouse=True)
    def setup_capybara(self):
        capybara.automatic_reload = False

    def test_reloads_the_current_context_of_the_node(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me")
        session.click_link("Reload!")
        sleep(0.3)
        assert node.reload().text == "has been reloaded"
        assert node.text == "has been reloaded"

    def test_reloads_a_parent_node(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me").find("css", "em")
        session.click_link("Reload!")
        sleep(0.3)
        assert node.reload().text == "has been reloaded"
        assert node.text == "has been reloaded"

    def test_does_not_automatically_reload(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me")
        session.click_link("Reload!")
        sleep(0.3)
        with pytest.raises(Exception) as excinfo:
            assert node.has_text("has been reloaded")
        assert isinstance(excinfo.value, session.driver.invalid_element_errors)


@pytest.mark.requires("js")
class TestNodeReloadWithAutomaticReload(NodeTestCase):
    @pytest.fixture(autouse=True)
    def setup_capybara(self):
        capybara.automatic_reload = True

    def test_reloads_the_current_context_of_the_node_automatically(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me")
        session.click_link("Reload!")
        sleep(0.3)
        assert node.text == "has been reloaded"

    def test_reloads_a_parent_node_automatically(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me").find("css", "em")
        session.click_link("Reload!")
        sleep(0.3)
        assert node.text == "has been reloaded"

    def test_reloads_a_node_automatically_when_using_find(self, session):
        session.visit("/with_js")
        node = session.find("css", "#reload-me")
        session.click_link("Reload!")
        sleep(0.3)
        assert node.find("css", "a").text == "has been reloaded"

    def test_does_not_reload_nodes_which_have_not_been_found_with_reevaluatable_queries(self, session):
        session.visit("/with_js")
        node = session.find_all("css", "#the-list li")[1]
        session.click_link("Fetch new list!")
        sleep(0.3)
        with pytest.raises(Exception) as excinfo:
            assert node.has_text("Foo")
        assert isinstance(excinfo.value, session.driver.invalid_element_errors)
        with pytest.raises(Exception) as excinfo:
            assert node.has_text("Bar")
        assert isinstance(excinfo.value, session.driver.invalid_element_errors)

    def test_reloads_nodes_with_options(self, session):
        session.visit("/with_js")
        node = session.find("css", "em", text="reloaded")
        session.click_link("Reload!")
        sleep(1)
        assert node.text == "has been reloaded"



import pytest

from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class ClickButtonTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestClickButton(ClickButtonTestCase):
    @pytest.mark.requires("js")
    def test_waits_for_asynchronous_load(self, session):
        session.visit("/with_js")
        session.click_link("Click me")
        session.click_button("New Here")

    def test_submits_the_latest_given_value_with_multiple_values_of_the_same_name(self, session):
        session.check("Terms of Use")
        session.click_button("awesome")
        assert extract_results(session).getlist("form[terms_of_use]")[-1] == "1"

    def test_submits_form_for_submit_button_by_id(self, session):
        session.click_button("awe123")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_submit_button_by_title(self, session):
        session.click_button("What an Awesome Button")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_submit_button_by_partial_title(self, session):
        session.click_button("What an Awesome")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_raises_an_error_for_partial_title_when_exact_is_true(self, session):
        with pytest.raises(ElementNotFound):
            session.click_button("What an Awesome", exact=True)

    def test_submits_form1_for_associated_button_located_within_form2(self, session):
        session.click_button("other_form_button")
        assert extract_results(session)["form[which_form]"] == "form1"

    def test_does_not_error_when_submit_button_not_associated_with_any_form_is_clicked(self, session):
        session.click_button("no_form_button")

    def test_submits_form_for_image_button_with_alt(self, session):
        session.click_button("oh hai thar")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_partial_alt(self, session):
        session.click_button("hai")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_value(self, session):
        session.click_button("okay")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_partial_value(self, session):
        session.click_button("kay")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_id(self, session):
        session.click_button("okay556")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_title(self, session):
        session.click_button("Okay 556 Image")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_image_button_with_partial_title(self, session):
        session.click_button("Okay 556")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_text(self, session):
        session.click_button("Click me")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_partial_text(self, session):
        session.click_button("Click")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_id(self, session):
        session.click_button("click_me_123")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_value(self, session):
        session.click_button("click_me")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_partial_value(self, session):
        session.click_button("ck_me")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_title(self, session):
        session.click_button("Click Title button")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_partial_title(self, session):
        session.click_button("Click Title")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_descendant_image_alt(self, session):
        session.click_button("A horse eating hay")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_submits_form_for_button_tag_with_partial_descendant_image_alt(self, session):
        session.click_button("se eating h")
        assert extract_results(session)["form[first_name]"] == "John"

    def test_raises_an_error_for_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.click_button("does not exist")
        assert "Unable to find button 'does not exist'" in str(excinfo.value)

    def test_serializes_and_sends_valueless_buttons_that_were_clicked(self, session):
        session.click_button("No Value!")
        assert extract_results(session)["form[no_value]"] is not None

    def test_does_not_send_image_buttons_that_were_not_clicked(self, session):
        session.click_button("awesome")
        assert extract_results(session).get("form[okay]") is None

    def test_raises_an_error_for_disabled_buttons(self, session):
        with pytest.raises(ElementNotFound):
            session.click_button("Disabled button")

    def test_encodes_complex_field_names(self, session):
        session.fill_in("address1_city", value="Paris")
        session.fill_in("address1_street", value="CDG")

        session.fill_in("address2_city", value="Mikolaiv")
        session.fill_in("address2_street", value="PGS")

        session.click_button("awesome")

        results = extract_results(session)

        assert results.getlist("form[addresses][][street]") == ["CDG", "PGS"]
        assert results.getlist("form[addresses][][city]") == ["Paris", "Mikolaiv"]
        assert results.getlist("form[addresses][][country]") == ["France", "Ukraine"]


class TestClickSubmitButtonWithHTML5Fields(ClickButtonTestCase):
    @pytest.fixture
    def results(self, session):
        session.click_button("html5_submit")
        return extract_results(session)

    def test_serializes_and_submits_search_fields(self, results):
        assert results["form[html5_search]"] == "what are you looking for"

    def test_serializes_and_submits_email_fields(self, results):
        assert results["form[html5_email]"] == "person@email.com"

    def test_serializes_and_submits_url_fields(self, results):
        assert results["form[html5_url]"] == "http://www.example.com"

    def test_serializes_and_submits_tel_fields(self, results):
        assert results["form[html5_tel]"] == "911"

    def test_serializes_and_submits_color_fields(self, results):
        assert results["form[html5_color]"].upper() == "#FFFFFF"


class TestClickSubmitButtonWithHTML4Fields(ClickButtonTestCase):
    @pytest.fixture
    def results(self, session):
        session.click_button("awesome")
        return extract_results(session)

    def test_serializes_and_submits_text_fields(self, results):
        assert results["form[first_name]"] == "John"

    def test_escapes_fields_when_submitting(self, results):
        assert results["form[phone]"] == "+1 555 7021"

    def test_serializes_and_submits_password_fields(self, results):
        assert results["form[password]"] == "seeekrit"

    def test_serializes_and_submits_hidden_fields(self, results):
        assert results["form[token]"] == "12345"

    def test_does_not_serialize_fields_from_other_forms(self, results):
        assert results.get("form[middle_name]") is None

    def test_submits_the_button_that_was_clicked_but_not_other_buttons(self, results):
        assert results["form[awesome]"] == "awesome"
        assert results.get("form[crappy]") is None

    def test_serializes_radio_buttons(self, results):
        assert results["form[gender]"] == "female"

    def test_defaults_radio_values_to_on_if_none_specified(self, results):
        assert results["form[valueless_radio]"] == "on"

    def test_serializes_check_boxes(self, results):
        assert "dog" in results.getlist("form[pets][]")
        assert "hamster" in results.getlist("form[pets][]")
        assert "cat" not in results.getlist("form[pets][]")

    def test_defaults_checkbox_value_to_on_if_none_specified(self, results):
        assert results["form[valueless_checkbox]"] == "on"

    def test_serializes_text_areas(self, results):
        assert results["form[description]"] == "Descriptive text goes here"

    def test_serializes_select_tag_with_values(self, results):
        assert results["form[locale]"] == "en"

    def test_serializes_select_tag_without_values(self, results):
        assert results["form[region]"] == "Norway"

    def test_serializes_first_option_for_select_tag_with_no_selection(self, results):
        assert results["form[city]"] == "London"

    def test_does_not_serialize_a_select_tag_without_options(self, results):
        assert results.get("form[tendency]") is None

    def test_converts_lf_to_cr_lf_in_submitted_textareas(self, results):
        assert results["form[newline]"] == "\r\nNew line after and before textarea tag\r\n"

    def test_does_not_submit_disabled_fields(self, results):
        assert results.get("form[disabled_text_field]") is None
        assert results.get("form[disabled_textarea]") is None
        assert results.get("form[disabled_checkbox]") is None
        assert results.get("form[disabled_radio]") is None
        assert results.get("form[disabled_select]") is None
        assert results.get("form[disabled_file]") is None


class TestClickButtonWithAssociatedFormFields(ClickButtonTestCase):
    @pytest.fixture
    def results(self, session):
        session.click_button("submit_form1")
        return extract_results(session)

    def test_serializes_and_submits_text_fields(self, results):
        assert results["form[outside_input]"] == "outside_input"

    def test_serializes_text_areas(self, results):
        assert results["form[outside_textarea]"] == "Some text here"

    def test_serializes_select_tags(self, results):
        assert results["form[outside_select]"] == "Ruby"

    def test_does_not_serialize_fields_associated_with_a_different_form(self, results):
        assert results.get("form[for_form2]") is None


class TestClickButtonTagOutsideForm(ClickButtonTestCase):
    @pytest.fixture
    def results(self, session):
        session.click_button("outside_button")
        return extract_results(session)

    def test_submits_the_associated_form(self, results):
        assert results["form[which_form]"] == "form2"

    def test_submits_the_button_that_was_clicked_but_not_other_buttons(self, results):
        assert results["form[outside_button]"] == "outside_button"
        assert results.get("form[unused]") is None


class TestClickSubmitTypeInputTagOutsideForm(ClickButtonTestCase):
    @pytest.fixture
    def results(self, session):
        session.click_button("outside_submit")
        return extract_results(session)

    def test_submits_the_associated_form(self, results):
        assert results["form[which_form]"] == "form1"

    def test_submits_the_button_that_was_clicked_but_not_other_buttons(self, results):
        assert results["form[outside_submit]"] == "outside_submit"
        assert results.get("form[submit_form1]") is None



from xpath import dsl as x
from xpath.renderer import to_xpath

from capybara.driver.base import Base
from capybara.werkzeug.browser import Browser
from capybara.werkzeug.node import Node


class Driver(Base):
    redirect_limit = 5

    def __init__(self, app):
        self.app = app
        self._browser = None

    @property
    def browser(self):
        if self._browser is None:
            self._browser = Browser(self)
        return self._browser

    @property
    def current_url(self):
        return self.browser.current_url

    @property
    def title(self):
        elements = self.browser.dom.xpath("/html/head/title")
        return elements[0].text if len(elements) else None

    @property
    def html(self):
        return self.browser.html

    def visit(self, path):
        self.browser.visit(path)

    def follow(self, method, path, params=None):
        self.browser.follow(method, path, params)

    def submit(self, method, path, params):
        self.browser.submit(method, path, params)

    def reset(self):
        self._browser = None

    def _find_css(self, css):
        elements = self.browser.dom.xpath(to_xpath(x.css(css)))
        return [Node(self, element) for element in elements]

    def _find_xpath(self, xpath):
        elements = self.browser.dom.xpath(xpath)
        return [Node(self, element) for element in elements]



from contextlib import closing
from threading import Thread

import capybara
from capybara.compat import URLError, urlopen, wsgi_decode_body, wsgi_encode_body
from capybara.utils import cached_property, find_available_port, timeout


class Server(object):
    """
    Serves a WSGI-compliant app for Capybara to test.

    Args:
        app (object): The WSGI-compliant app to serve.
        port (int, optional): The port on which the server should be available.
            Defaults to :data:`capybara.server_port`, or the last port used by the
            given app, or a random available port.
        host (str, optional): The host on which the server should be available.
            Defaults to :data:`capybara.server_host`.
    """

    _ports = {}

    def __init__(self, app, port=None, host=None):
        self.app = app

        self.host = (
            host or
            capybara.server_host)

        self.port = (
            port or
            capybara.server_port or
            type(self)._ports.get(self.port_key, None) or
            find_available_port())

        self.server_thread = None

    @property
    def port_key(self):
        return str(id(self.app))

    @cached_property
    def middleware(self):
        return Middleware(self.app)

    def boot(self):
        """
        Boots a server for the app, if it isn't already booted.

        Returns:
            Server: This server.
        """

        if not self.responsive:
            # Remember the port so we can reuse it if we try to serve this same app again.
            type(self)._ports[self.port_key] = self.port

            init_func = capybara.servers[capybara.server_name]
            init_args = (self.middleware, self.port, self.host)

            self.server_thread = Thread(target=init_func, args=init_args)

            # Inform Python that it shouldn't wait for this thread to terminate before
            # exiting. (It will still be appropriately terminated when the process exits.)
            self.server_thread.daemon = True

            self.server_thread.start()

            # Make sure the server actually starts and becomes responsive.
            with timeout(60):
                while not self.responsive:
                    self.server_thread.join(0.1)

        return self

    @property
    def responsive(self):
        """ bool: Whether the server for this app is up and responsive. """

        if self.server_thread and self.server_thread.join(0):
            return False

        try:
            # Try to fetch the endpoint added by the middleware.
            identify_url = "http://{0}:{1}/__identify__".format(self.host, self.port)

            with closing(urlopen(identify_url)) as response:
                body, status_code = response.read(), response.getcode()

                if str(status_code)[0] == "2" or str(status_code)[0] == "3":
                    return wsgi_decode_body(body) == str(id(self.app))
        except URLError:
            pass

        return False


class Middleware(object):
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        if environ["PATH_INFO"] == "/__identify__":
            return self.identify(environ, start_response)
        else:
            return self.app(environ, start_response)

    def identify(self, environ, start_response):
        start_response("200 OK", [("Content-Type", "text/plain")])
        return [wsgi_encode_body(str(id(self.app)))]



import re

from capybara.utils import decode_bytes, isbytes, isregex


def declension(singular, plural, count):
    """
    Returns the appropriate word variation for the given quantity.

    Args:
        singular (str): The singular variation.
        plural (str): The plural variation.
        count (int): The count.

    Returns:
        str: The appropriate variation for the quantity.
    """

    return singular if count == 1 else plural


def desc(value):
    """ str | List[str]: A normalized representation for a user-provided value. """

    if isinstance(value, list):
        return map(desc, value)

    if isregex(value):
        value = value.pattern

    if isbytes(value):
        value = decode_bytes(value)

    return repr(value)


def expects_none(options):
    """
    Returns whether the given query options expect a possible count of zero.

    Args:
        options (Dict[str, int | Iterable[int]]): A dictionary of query options.

    Returns:
        bool: Whether a possible count of zero is expected.
    """

    if any(options.get(key) is not None for key in ["count", "maximum", "minimum", "between"]):
        return matches_count(0, options)
    else:
        return False


def failure_message(description, options):
    """
    Returns a expectation failure message for the given query description.

    Args:
        description (str): A description of the failed query.
        options (Dict[str, Any]): The query options.

    Returns:
        str: A message describing the failure.
    """

    message = "expected to find {}".format(description)

    if options["count"] is not None:
        message += " {count} {times}".format(
            count=options["count"],
            times=declension("time", "times", options["count"]))
    elif options["between"] is not None:
        between = options["between"]
        if between:
            first, last = between[0], between[-1]
        else:
            first, last = None, None

        message += " between {first} and {last} times".format(
            first=first,
            last=last)
    elif options["maximum"] is not None:
        message += " at most {maximum} {times}".format(
            maximum=options["maximum"],
            times=declension("time", "times", options["maximum"]))
    elif options["minimum"] is not None:
        message += " at least {minimum} {times}".format(
            minimum=options["minimum"],
            times=declension("time", "times", options["minimum"]))

    return message


def matches_count(count, options):
    """
    Returns whether the given count matches the given query options.

    If no quantity options are specified, any count is considered acceptable.

    Args:
        count (int): The count to be validated.
        options (Dict[str, int | Iterable[int]]): A dictionary of query options.

    Returns:
        bool: Whether the count matches the options.
    """

    if options.get("count") is not None:
        return count == int(options["count"])
    if options.get("maximum") is not None and int(options["maximum"]) < count:
        return False
    if options.get("minimum") is not None and int(options["minimum"]) > count:
        return False
    if options.get("between") is not None and count not in options["between"]:
        return False
    return True


def normalize_text(value):
    """
    Normalizes the given value to a string of text with extra whitespace removed.

    Byte sequences are decoded. ``None`` is converted to an empty string. Everything else
    is simply cast to a string.

    Args:
        value (Any): The data to normalize.

    Returns:
        str: The normalized text.
    """

    if value is None:
        return ""

    text = decode_bytes(value) if isbytes(value) else str(value)

    return normalize_whitespace(text)


def normalize_whitespace(text):
    """
    Returns the given text with outer whitespace removed and inner whitespace collapsed.

    Args:
        text (str): The text to normalize.

    Returns:
        str: The normalized text.
    """

    return re.sub(r"\s+", " ", text, flags=re.UNICODE).strip()


def toregex(text):
    """
    Returns a compiled regular expression for the given text.

    Args:
        text (str | RegexObject): The text to match.

    Returns:
        RegexObject: A compiled regular expression that will match the text.
    """

    return (text if isregex(text)
            else re.compile(re.escape(normalize_text(text))))



import pytest


class TestReset:
    def test_removes_cookies(self, session):
        session.visit("/set_cookie")
        session.visit("/get_cookie")
        assert session.has_text("test_cookie")

        session.reset()
        session.visit("/get_cookie")
        assert "test_cookie" not in session.body

    def test_resets_current_url_host_and_path(self, session):
        session.visit("/foo")
        assert session.current_url
        assert session.current_url
        assert session.current_path == "/foo"

        session.reset()
        assert session.current_url in [None, "", "about:blank"]
        assert session.current_path in [None, "", "blank"]
        assert session.current_host is None

    def test_resets_page_body(self, session):
        session.visit("/with_html")
        assert session.has_text("This is a test")
        assert "This is a test" in session.find(".//h1").text

        session.reset()
        assert "This is a test" not in session.body
        assert session.has_no_selector(".//h1")

    def test_is_synchronous(self, session):
        session.visit("/with_slow_unload")
        assert session.has_selector("css", "div")
        session.reset()
        assert session.has_no_selector("xpath", "/html/body/*", wait=False)

    @pytest.mark.requires("modals")
    def test_handles_modals_during_unload(self, session):
        session.visit("/with_unload_alert")
        assert session.has_selector("css", "div")
        session.reset()
        assert session.has_no_selector("xpath", "/html/body/*", wait=False)

    @pytest.mark.requires("modals")
    def test_handles_already_open_modals(self, session):
        session.visit("/with_unload_alert")
        session.click_link("Go away")
        session.reset()
        assert session.has_no_selector("xpath", "/html/body/*", wait=False)



from functools import wraps
import sys

import capybara
from capybara import DSL_METHODS as PACKAGE_METHODS
from capybara.session import DSL_METHODS as SESSION_METHODS, Session


__all__ = ["page"] + SESSION_METHODS + PACKAGE_METHODS


class Page(object):
    """ A proxy for the current session. """
    def __getattr__(self, attr):
        return getattr(capybara.current_session(), attr)


page = Page()
""" The singleton current-session proxy object. """


class DSLMixin:
    """ A mixin for including DSL methods in another class. """
    pass


_module_name = globals()["__name__"]
_module = sys.modules[_module_name]


def _define_package_method(name):
    func = getattr(capybara, name)

    setattr(DSLMixin, name, func)
    setattr(_module, name, func)


def _define_session_method(name):
    @wraps(getattr(Session, name))
    def func(*args, **kwargs):
        return getattr(page, name)(*args, **kwargs)

    setattr(DSLMixin, name, func)
    setattr(_module, name, func)


for _method in PACKAGE_METHODS:
    _define_package_method(_method)


for _method in SESSION_METHODS:
    _define_session_method(_method)



import pytest


@pytest.mark.requires("js")
class TestGoBack:
    def test_fetches_a_response_from_the_driver_from_the_previous_page(self, session):
        session.visit("/")
        assert session.has_text("Hello world!")
        session.visit("/foo")
        assert session.has_text("Another World")
        session.go_back()
        assert session.has_text("Hello world!")



import os.path

import capybara
from capybara.exceptions import ElementNotFound, FileNotFound


class ActionsMixin(object):
    """
    If the driver is capable of executing JavaScript, actions will wait for a set amount of time and
    continuously retry finding the element until either the element is found or the time expires.
    The length of time :meth:`find` will wait is controlled through
    :data:`capybara.default_max_wait_time`.
    """

    def attach_file(self, locator, path, **kwargs):
        """
        Find a file field on the page and attach a file given its path. The file field can be found
        via its name, id, or label text. ::

            page.attach_file(locator, "/path/to/file.png")

        Args:
            locator (str): Which field to attach the file to.
            path (str): The path of the file that will be attached.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.

        Raises:
            FileNotFound: No file exists at the given path.
        """

        if not os.path.isfile(path):
            raise FileNotFound("cannot attach file, {0} does not exist".format(path))

        self.find("file_field", locator, **kwargs).set(path)

    def check(self, locator, allow_label_click=None, **kwargs):
        """
        Find a check box and mark it as checked. The check box can be found via name, id, or label
        text. ::

            page.check("German")

        Args:
            locator (str): Which check box to check.
            allow_label_click (bool, optional): Attempt to click the label to toggle state if
                element is non-visible. Defaults to :data:`capybara.automatic_label_click`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        if allow_label_click is None:
            allow_label_click = capybara.automatic_label_click

        try:
            self.find("checkbox", locator, **kwargs).set(True)
        except ElementNotFound as e:
            if not allow_label_click:
                raise
            try:
                kwargs["visible"] = "hidden"
                kwargs["wait"] = False
                checkbox = self.find("checkbox", locator, **kwargs)
                label = self.find("label", field=checkbox, visible=True, wait=False)
                if not checkbox.checked:
                    label.click()
            except Exception:
                raise e

    def choose(self, locator, allow_label_click=None, **kwargs):
        """
        Find a radio button and mark it as checked. The radio button can be found via name, id, or
        label text. ::

            page.choose("Male")

        Args:
            locator (str): Which radio button to choose.
            allow_label_click (bool, optional): Attempt to click the label to toggle state if
                element is non-visible. Defaults to :data:`capybara.automatic_label_click`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        if allow_label_click is None:
            allow_label_click = capybara.automatic_label_click

        try:
            self.find("radio_button", locator, **kwargs).set(True)
        except ElementNotFound as e:
            if not allow_label_click:
                raise
            try:
                kwargs["visible"] = "hidden"
                kwargs["wait"] = False
                radio_button = self.find("radio_button", locator, **kwargs)
                label = self.find("label", field=radio_button, visible=True, wait=False)
                if not radio_button.checked:
                    label.click()
            except Exception:
                raise e

    def click_button(self, locator, **kwargs):
        """
        Finds a button on the page and clicks it. This can be any ``<input>`` element of type
        submit, reset, image, or button, or it can be any ``<button>`` element. All buttons can be
        found by their id, value, or title. ``<button>`` elements can also be found by their text
        content, and image ``<input>`` elements by their alt attribute.

        Args:
            locator (str): Which button to find.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        return self.find("button", locator, **kwargs).click()

    def click_link(self, locator, **kwargs):
        """
        Finds a link by id, text, or title and clicks it. Also looks at image alt text inside the
        link.

        Args:
            locator (str): Text, id, title, or nested image's alt attribute.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        return self.find("link", locator, **kwargs).click()

    def click_link_or_button(self, locator, **kwargs):
        """
        Finds a button or link by id, text or value and clicks it. Also looks at image alt text
        inside the link.

        Args:
            locator (str): Text, id, or value of link or button.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        return self.find("link_or_button", locator, **kwargs).click()

    click_on = click_link_or_button
    """ Alias for :meth:`click_link_or_button`. """

    def fill_in(self, locator, value=None, **kwargs):
        """
        Locate a text field or text area and fill it in with the given text. The field can be found
        via its name, id, or label text. ::

            page.fill_in("Name", value="Bob")

        Args:
            locator (str): Which field to fill in.
            value (str, optional): The value to fill in. Defaults to None.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        self.find("fillable_field", locator, **kwargs).set(value)

    def select(self, value, field=None, **kwargs):
        """
        If the ``field`` argument is present, ``select`` finds a select box on the page and selects
        a particular option from it. Otherwise it finds an option inside the current scope and
        selects it. If the select box is a multiple select, ``select`` can be called multiple times
        to select more than one option. The select box can be found via its name, id, or label text.
        The option can be found by its text. ::

            page.select("March", field="Month")

        Args:
            value (str): Which option to select.
            field (str, optional): The id, name, or label of the select box.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        if field:
            self.find("select", field, **kwargs).find("option", value, **kwargs).select_option()
        else:
            self.find("option", value, **kwargs).select_option()

    def uncheck(self, locator, allow_label_click=None, **kwargs):
        """
        Find a check box and uncheck it. The check box can be found via name, id, or label text. ::

            page.uncheck("German")

        Args:
            locator (str): Which check box to uncheck.
            allow_label_click (bool, optional): Attempt to click the label to toggle state if
                element is non-visible. Defaults to :data:`capybara.automatic_label_click`.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        if allow_label_click is None:
            allow_label_click = capybara.automatic_label_click

        try:
            self.find("checkbox", locator, **kwargs).set(False)
        except ElementNotFound as e:
            if not allow_label_click:
                raise
            try:
                kwargs["visible"] = "hidden"
                kwargs["wait"] = False
                checkbox = self.find("checkbox", locator, **kwargs)
                label = self.find("label", field=checkbox, visible=True, wait=False)
                if checkbox.checked:
                    label.click()
            except Exception:
                raise e

    def unselect(self, value, field=None, **kwargs):
        """
        Find a select box on the page and unselect a particular option from it. If the select box is
        a multiple select, ``unselect`` can be called multiple times to unselect more than one
        option. The select box can be found via its name, id, or label text. ::

            page.unselect("March", field="Month")

        Args:
            value (str): Which option to unselect.
            field (str, optional): The id, name, or label of the select box.
            **kwargs: Arbitrary keyword arguments for :class:`SelectorQuery`.
        """

        if field:
            self.find("select", field, **kwargs).find("option", value, **kwargs).unselect_option()
        else:
            self.find("option", value, **kwargs).unselect_option()



import pytest
import re


class TestHasLink:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_true_if_the_given_link_is_on_the_page(self, session):
        assert session.has_link("foo")
        assert session.has_link("awesome title")
        assert session.has_link("A link", href="/with_simple_html")
        assert session.has_link("A link", href=re.compile(r"\/with_simple_html"))

    def test_is_false_if_the_given_link_is_not_on_the_page(self, session):
        assert not session.has_link("monkey")
        assert not session.has_link("A link", href="/non-existant-href")
        assert not session.has_link("A link", href=re.compile(r"non-existant"))


class TestHasNoLink:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_html")

    def test_is_false_if_the_given_link_is_on_the_page(self, session):
        assert not session.has_no_link("foo")
        assert not session.has_no_link("awesome title")
        assert not session.has_no_link("A link", href="/with_simple_html")
        assert not session.has_no_link("A link", href=re.compile(r"\/with_simple_html"))

    def test_is_true_if_the_given_link_is_not_on_the_page(self, session):
        assert session.has_no_link("monkey")
        assert session.has_no_link("A link", href="/non-existant-href")
        assert session.has_no_link("A link", href=re.compile(r"non-existant"))



# coding=utf-8
from __future__ import unicode_literals

import capybara


class TestText:
    def test_returns_the_text_of_the_page(self, session):
        session.visit("/with_simple_html")
        assert session.text == "Bar"

    def test_ignores_invisible_text_by_default(self, session):
        session.visit("/with_html")
        assert session.find("xpath", "//*[@id='hidden-text']").text == "Some of this text is"

    def test_ignores_invisible_text_if_ignore_hidden_elements_is_true(self, session):
        session.visit("/with_html")
        assert session.find("id", "hidden-text").text == "Some of this text is"
        capybara.ignore_hidden_elements = False
        assert session.find("id", "hidden-text").text == "Some of this text is hidden!"

    def test_ignores_invisible_text_if_visible_text_only_is_true(self, session):
        session.visit("/with_html")
        capybara.visible_text_only = True
        assert session.find("id", "hidden-text").text == "Some of this text is"
        capybara.ignore_hidden_elements = False
        assert session.find("id", "hidden-text").text == "Some of this text is"

    def test_ignores_invisible_text_if_ancestor_is_invisible(self, session):
        session.visit("/with_html")
        assert session.find("id", "hidden_via_ancestor", visible=False).text == ""

    def test_strips_whitespace(self, session):
        session.visit("/with_html")
        assert "text with whitespace" in session.find("xpath", "//*[@id='second']").text

    def test_returns_unicode_text(self, session):
        session.visit("/with_html")
        assert session.find("id", "unicode-text").text == "이름"


class TestAllText:
    def test_shows_invisible_text(self, session):
        session.visit("/with_html")
        assert session.find("id", "hidden-text").all_text == "Some of this text is hidden!"


class TestVisibleText:
    def test_ignores_invisible_text(self, session):
        session.visit("/with_html")
        assert session.find("id", "hidden-text").visible_text == "Some of this text is"



import pytest

import capybara
from capybara.exceptions import ElementNotFound
from capybara.tests.helpers import extract_results


class CheckTestCase:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/form")


class TestCheck(CheckTestCase):
    def test_checked_attribute_s_true_if_checked(self, session):
        session.check("Terms of Use")
        assert session.find("xpath", "//input[@id='form_terms_of_use']").checked

    def test_checked_attribute_is_false_if_unchecked(self, session):
        assert not session.find("xpath", "//input[@id='form_terms_of_use']").checked

    @pytest.mark.requires("js")
    def test_triggers_associated_events(self, session):
        session.visit("/with_js")
        session.check("checkbox_with_event")
        assert session.has_css("#checkbox_event_triggered")

    def test_checking_does_not_change_an_already_checked_checkbox(self, session):
        assert session.find("xpath", "//input[@id='form_pets_dog']").checked
        session.check("form_pets_dog")
        assert session.find("xpath", "//input[@id='form_pets_dog']").checked

    def test_checking_checks_an_unchecked_checkbox(self, session):
        assert not session.find("xpath", "//input[@id='form_pets_cat']").checked
        session.check("form_pets_cat")
        assert session.find("xpath", "//input[@id='form_pets_cat']").checked

    def test_unchecking_does_not_change_an_already_unchecked_checkbox(self, session):
        assert not session.find("xpath", "//input[@id='form_pets_cat']").checked
        session.uncheck("form_pets_cat")
        assert not session.find("xpath", "//input[@id='form_pets_cat']").checked

    def test_unchecking_unchecks_a_checked_checkbox(self, session):
        assert session.find("xpath", "//input[@id='form_pets_dog']").checked
        session.uncheck("form_pets_dog")
        assert not session.find("xpath", "//input[@id='form_pets_dog']").checked

    def test_checks_a_checkbox_by_id(self, session):
        session.check("form_pets_cat")
        session.click_button("awesome")
        pets = extract_results(session).getlist("form[pets][]")
        assert "dog" in pets
        assert "cat" in pets
        assert "hamster" in pets

    def test_checks_a_checkbox_by_label(self, session):
        session.check("Cat")
        session.click_button("awesome")
        pets = extract_results(session).getlist("form[pets][]")
        assert "dog" in pets
        assert "cat" in pets
        assert "hamster" in pets

    def test_raises_an_error_for_a_locator_that_does_not_exist(self, session):
        with pytest.raises(ElementNotFound):
            session.check("does not exist")

    def test_raises_an_error_for_a_disabled_checkbox(self, session):
        with pytest.raises(ElementNotFound):
            session.check("Disabled Checkbox")


class TestCheckWithAutomaticLabelClick(CheckTestCase):
    @pytest.fixture(autouse=True)
    def setup_settings(self):
        old_automatic_label_click = capybara.automatic_label_click
        capybara.automatic_label_click = True
        try:
            yield
        finally:
            capybara.automatic_label_click = old_automatic_label_click

    def test_checks_via_clicking_the_label_with_for_attribute_if_possible(self, session):
        assert session.find("checkbox", "form_cars_tesla", unchecked=True, visible="hidden")
        session.check("form_cars_tesla")
        session.click_button("awesome")
        assert "tesla" in extract_results(session).getlist("form[cars][]")

    def test_checks_via_clicking_the_wrapping_label_if_possible(self, session):
        assert session.find("checkbox", "form_cars_mclaren", unchecked=True, visible="hidden")
        session.check("form_cars_mclaren")
        session.click_button("awesome")
        assert "mclaren" in extract_results(session).getlist("form[cars][]")

    def test_does_not_click_the_label_if_unneeded(self, session):
        assert session.find("checkbox", "form_cars_jaguar", checked=True, visible="hidden")
        session.check("form_cars_jaguar")
        session.click_button("awesome")
        assert "jaguar" in extract_results(session).getlist("form[cars][]")

    def test_raises_original_error_when_no_label_available(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.check("form_cars_ariel")
        assert "Unable to find checkbox 'form_cars_ariel'" in str(excinfo.value)

    def test_raises_error_if_not_allowed_to_click_label(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.check("form_cars_mclaren", allow_label_click=False)
        assert "Unable to find checkbox 'form_cars_mclaren'" in str(excinfo.value)


class TestCheckWithoutAutomaticLabelClick(CheckTestCase):
    @pytest.fixture(autouse=True)
    def setup_settings(self):
        old_automatic_label_click = capybara.automatic_label_click
        capybara.automatic_label_click = False
        try:
            yield
        finally:
            capybara.automatic_label_click = old_automatic_label_click

    def test_raises_error_if_checkbox_not_visible(self, session):
        with pytest.raises(ElementNotFound) as excinfo:
            session.check("form_cars_mclaren")
        assert "Unable to find checkbox 'form_cars_mclaren'" in str(excinfo.value)

    def test_checks_via_the_label_if_allow_label_click_is_true(self, session):
        assert session.find("checkbox", "form_cars_tesla", unchecked=True, visible="hidden")
        session.check("form_cars_tesla", allow_label_click=True)
        session.click_button("awesome")
        assert "tesla" in extract_results(session).getlist("form[cars][]")



import pytest

from capybara.exceptions import ModalNotFound


@pytest.mark.requires("modals")
class TestAcceptAlert:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_accepts_the_alert(self, session):
        with session.accept_alert():
            session.click_link("Open alert")
        assert session.has_xpath("//a[@id='open-alert' and @opened='true']")

    def test_accepts_the_alert_if_the_text_matches(self, session):
        with session.accept_alert("Alert opened"):
            session.click_link("Open alert")
        assert session.has_xpath("//a[@id='open-alert' and @opened='true']")

    def test_raises_an_error_if_the_text_does_not_match(self, session):
        with pytest.raises(ModalNotFound):
            with session.accept_alert("Incorrect Text"):
                session.click_link("Open alert")

    def test_accepts_an_asynchronous_alert(self, session):
        with session.accept_alert():
            session.click_link("Open delayed alert")
        assert session.has_xpath("//a[@id='open-delayed-alert' and @opened='true']")

    def test_allows_adjustment_of_the_delay(self, session):
        with session.accept_alert(wait=4):
            session.click_link("Open slow alert")
        assert session.has_xpath("//a[@id='open-slow-alert' and @opened='true']")



import pytest

from capybara.exceptions import ModalNotFound


@pytest.mark.requires("modals")
class TestDismissConfirm:
    @pytest.fixture(autouse=True)
    def setup_session(self, session):
        session.visit("/with_js")

    def test_dismisses_the_confirm(self, session):
        with session.dismiss_confirm():
            session.click_link("Open confirm")
        assert session.has_xpath("//a[@id='open-confirm' and @confirmed='false']")

    def test_dismisses_the_confirm_if_the_message_matches(self, session):
        with session.dismiss_confirm("Confirm opened"):
            session.click_link("Open confirm")
        assert session.has_xpath("//a[@id='open-confirm' and @confirmed='false']")

    def test_raises_an_error_if_the_message_does_not_match(self, session):
        with pytest.raises(ModalNotFound):
            with session.dismiss_confirm("Incorrect Text"):
                session.click_link("Open confirm")



from capybara.exceptions import ExpectationNotMet
from capybara.queries.current_path_query import CurrentPathQuery


class SessionMatchersMixin(object):
    def assert_current_path(self, path, **kwargs):
        """
        Asserts that the page has the given path. By default this will compare against the
        path+query portion of the full URL.

        Args:
            path (str | RegexObject): The string or regex that the current "path" should match.
            **kwargs: Arbitrary keyword arguments for :class:`CurrentPathQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = CurrentPathQuery(path, **kwargs)

        @self.document.synchronize
        def assert_current_path():
            if not query.resolves_for(self):
                raise ExpectationNotMet(query.failure_message)
        assert_current_path()

        return True

    def assert_no_current_path(self, path, **kwargs):
        """
        Asserts that the page doesn't have the given path.

        Args:
            path (str | RegexObject): The string or regex that the current "path" should match.
            **kwargs: Arbitrary keyword arguments for :class:`CurrentPathQuery`.

        Returns:
            True

        Raises:
            ExpectationNotMet: If the assertion hasn't succeeded during the wait time.
        """

        query = CurrentPathQuery(path, **kwargs)

        @self.document.synchronize
        def assert_no_current_path():
            if query.resolves_for(self):
                raise ExpectationNotMet(query.negative_failure_message)

        assert_no_current_path()

        return True

    def has_current_path(self, path, **kwargs):
        """
        Checks if the page has the given path.

        Args:
            path (str | RegexObject): The string or regex that the current "path" should match.
            **kwargs: Arbitrary keyword arguments for :class:`CurrentPathQuery`.

        Returns:
            bool: Whether it matches.
        """

        try:
            return self.assert_current_path(path, **kwargs)
        except ExpectationNotMet:
            return False

    def has_no_current_path(self, path, **kwargs):
        """
        Checks if the page doesn't have the given path.

        Args:
            path (str | RegexObject): The string or regex that the current "path" should match.
            **kwargs: Arbitrary keyword arguments for :class:`CurrentPathQuery`.

        Returns:
            bool: Whether it doesn't match.
        """

        try:
            return self.assert_no_current_path(path, **kwargs)
        except ExpectationNotMet:
            return False
